package de.kreawi.trainer;

import java.beans.PropertyVetoException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.commons.io.FileUtils;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.de.GermanAnalyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field.Store;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.IndexWriterConfig;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.IOContext;
import org.apache.lucene.store.RAMDirectory;
import org.h2.tools.DeleteDbFiles;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jquizengine.questions.Question;
import com.jquizengine.questions.collection.QuestionPool;
import com.jquizengine.questions.readerwriter.InvalidReaderWriterException;
import com.jquizengine.questions.readerwriter.QuestionReadWriteException;
import com.jquizengine.questions.readerwriter.QuestionReader;
import com.jquizengine.questions.readerwriter.ReaderWriter;
import com.jquizengine.questions.readerwriter.ReaderWriterFactory;
import com.jquizengine.quiz.Quiz;
import com.mchange.v2.c3p0.ComboPooledDataSource;
import com.mchange.v2.c3p0.DataSources;

import de.kreawi.trainer.db.QuestionAttributes;
import de.kreawi.trainer.db.QuestionAttributesModel;
import de.kreawi.trainer.db.TrainerOnlineGamePersistence;
import de.kreawi.trainer.db.TrainerSettingsModel;
import de.kreawi.trainer.dialogs.backup.TrainerBackup;
import de.kreawi.trainer.dialogs.history.HistoryItem;
import de.kreawi.trainer.exceptions.TrainerStorageException;
import de.kreawi.trainer.onlinegame.OnlineGamePersistence;
import de.kreawi.trainer.quiz.KreawiQuiz;
import de.kreawi.trainer.quiz.QuizModel;
import de.kreawi.trainer.util.TrainerHtmlUtil;
import de.kreawi.trainer.util.TrainerTextResources;
import de.kreawi.trainer.util.TrainerUtil;
import de.martinknauer.util.config.ConfigFile;
import de.martinknauer.util.progress.ProgressChangeEvent;
import de.martinknauer.util.progress.ProgressListener;

public class TrainerStorage {
  
  private static final String INTERNAL_SETTINGS_FILE_PATH = "de/kreawi/trainer/resources/trainer_settings.xml";
  
  private static final String APP_DATA_HISTORY_DIR_NAME = "history";
  
  private static final String BACKUP_DIR_NAME = "backup";

  private static final String BACKUP_META_FILE_NAME = "backup_info";

  private static final String BACKUP_DB_FILE_NAME = "db.bak";
  
  private static final String TMP_DB_NAME = "tmp";

  private static final String HISTORY_FILE_SUFFIX = ".bin";
  
  private static final String HISTORY_FILE_SUFFIX_SEPARATOR = "-";

  private static final String HISTORY_MC_FILE_SUFFIX = HISTORY_FILE_SUFFIX_SEPARATOR
      + "mc" + HISTORY_FILE_SUFFIX;

  private static final String HISTORY_TEXT_FILE_SUFFIX = HISTORY_FILE_SUFFIX_SEPARATOR
      + "text" + HISTORY_FILE_SUFFIX;
  
  private static final String APP_DATA_SEARCH_INDEX_DIR_NAME = "search_index";

  private static final String SEARCH_INDEX_FIELD_ID = "id";

  private static final String SEARCH_INDEX_FIELD_CONTENTS = "searchableText";
  
  
  private ConfigFile config;
  
  private QuestionPool allQuestions;
  
  private List<HistoryItem> historyItemsText;

  private List<HistoryItem> historyItemsMc;

  private Directory searchIndexRepo;
  
  private TrainerSettingsModel settingsModel;
  private QuestionAttributesModel questionAttributesModel;
  private TrainerOnlineGamePersistence onlineGamePersistence;

  private ComboPooledDataSource dataSourcePooled;

  private ArrayList<McHistoryChangedListener> mcHistoryChangedListeners = new ArrayList<McHistoryChangedListener>();
  private ArrayList<TextHistoryChangedListener> textHistoryChangedListeners = new ArrayList<TextHistoryChangedListener>();
  
  private static final Logger logger = LoggerFactory.getLogger(TrainerStorage.class);

  public TrainerStorage() throws TrainerStorageException {
    
    initDb();
    initQuestionsAndHistory();
    
  }

  private void initDb() {

    final File dbFile = KreawiGlobals.getDbFile();
    logger.info("Opening db connection from db file: " + dbFile.getAbsolutePath());
    if (!dbFile.exists()) {
      logger.info("DB file doesn't exist. Creating a new one!");
    }
    
    try {
      ComboPooledDataSource cpds = new ComboPooledDataSource(); 
      cpds.setDriverClass(KreawiGlobals.DB_DRIVER_NAME); //loads the jdbc driver
      cpds.setJdbcUrl(KreawiGlobals.getDbUrl());
      cpds.setUser(KreawiGlobals.DB_USER);
      cpds.setPassword(KreawiGlobals.DB_PASSWORD);
      this.dataSourcePooled = cpds;

      this.settingsModel = new TrainerSettingsModel(this.dataSourcePooled);
      this.questionAttributesModel = new QuestionAttributesModel(this.dataSourcePooled);
      this.onlineGamePersistence = new TrainerOnlineGamePersistence(this.dataSourcePooled);
      
    } catch (PropertyVetoException e) {
      logger.error("DB connection pool could not be initialized!", e);
    }
    
  }

  private void initQuestionsAndHistory() throws TrainerStorageException {
    loadAllQuestions();
    loadHistoryItemsMc();
    loadHistoryItemsText();
    initSearchIndex();
  }

  public void closeDb() {
    try {
      this.dataSourcePooled.close();
      DataSources.destroy(this.dataSourcePooled);
    } catch (SQLException e) {
      logger.error("Error on closing data source!", e);
    }
  }

  private void initSearchIndex() throws TrainerStorageException {
    try {
      if (KreawiGlobals.isDemo()) {
        logger.info("Demo mode detected. Creating search index in memory...");
        createSearchIndex(getRAMIndexDir());
      } else {
        if (isSearchIndexCreationNecessary()) {
          logger.info("Search index update necessary. (Re)creating file system index...");
          createSearchIndex(getFSIndexDir());
          updateTimestamps();
        } else {
          logger.info("Search index is up to date. It will not be (re)created.");
        }
        // Initialize search index in memory (also automatically initialized on reading the index)
        getRAMIndexDir();    
      }
    } catch (IOException e) {
      logger.error("Fehler beim Erstellen des Fragen-Indexes!", e);
      throw new TrainerStorageException();
    }
  }
  
  private boolean isSearchIndexCreationNecessary() {
    final String timestampDataFile = this.settingsModel.getValue(TrainerSettingsConstants.KEY_TIMESTAMP_DATA_FILE);
    final String timestampUpdateFile = this.settingsModel.getValue(TrainerSettingsConstants.KEY_TIMESTAMP_UPDATE_FILE);
    final File workDir = new File(System.getProperty("user.dir"));
    final File dataFile = new File(new File(workDir, KreawiGlobals.QUESTION_DIR_NAME), KreawiGlobals.QUESTION_FILE_NAME);
    final File updateFile = new File(new File(workDir, KreawiGlobals.QUESTION_UPDATES_DIR_NAME), KreawiGlobals.QUESTION_UPDATES_FILE_NAME);
    
    if (dataFile.exists()) {
      if (!Long.toString(dataFile.lastModified()).equals(timestampDataFile)) {
        return true;
      }
    }
    if (updateFile.exists()) {
      if (!Long.toString(dataFile.lastModified()).equals(timestampUpdateFile)) {
        return true;
      }
    }
    
    return false;
  }

  private void updateTimestamps() {
    final File workDir = new File(System.getProperty("user.dir"));
    final File dataFile = new File(new File(workDir, KreawiGlobals.QUESTION_DIR_NAME), KreawiGlobals.QUESTION_FILE_NAME);
    final File updateFile = new File(new File(workDir, KreawiGlobals.QUESTION_UPDATES_DIR_NAME), KreawiGlobals.QUESTION_UPDATES_FILE_NAME);
    
    if (dataFile.exists()) {
      this.settingsModel.setLongValue(TrainerSettingsConstants.KEY_TIMESTAMP_DATA_FILE, dataFile.lastModified());
    }
    
    if (updateFile.exists()) {
      this.settingsModel.setLongValue(TrainerSettingsConstants.KEY_TIMESTAMP_UPDATE_FILE, updateFile.lastModified());
    }
  }

  private void createSearchIndex(final Directory indexDir) throws TrainerStorageException {
    
    try {
      final long startTime = System.currentTimeMillis();
      Analyzer analyzer = new StandardAnalyzer(GermanAnalyzer.getDefaultStopSet());
      IndexWriterConfig conf = new IndexWriterConfig(analyzer);
      IndexWriter indexWriter = new IndexWriter(indexDir, conf);
      for (Question question : allQuestions.getQuestions()) {
        Document document = new Document();
        
        document.add(new StoredField(SEARCH_INDEX_FIELD_ID, question.getId()));
        
        final String htmlText = "ID:" + question.getId() + " " + question.getText() + " " + question.getComment();
        final String searchableText = TrainerHtmlUtil.removeHtmlTags(htmlText);
        document.add(new TextField(SEARCH_INDEX_FIELD_CONTENTS, searchableText, Store.NO));
        
        indexWriter.addDocument(document);
      }
      indexWriter.close();
      if (KreawiGlobals.isDevMode() || KreawiGlobals.isBetaMode()) {
        logger.info("Laufzeit der Suchindex-Generierung: " + (System.currentTimeMillis() - startTime) + " ms");
      }
    } catch (IOException e) {
      logger.error("Fehler beim Erstellen des Fragen-Indexes!", e);
      throw new TrainerStorageException();
    }
  }

  private Directory getRAMIndexDir() throws IOException {
    if (this.searchIndexRepo == null) {
      if (KreawiGlobals.isDemo()) {
        this.searchIndexRepo = new RAMDirectory();
      } else {
        this.searchIndexRepo = new RAMDirectory(getFSIndexDir(), IOContext.DEFAULT);
      }
    }
    return searchIndexRepo;
  }
  
  private FSDirectory getFSIndexDir() throws IOException {
    return FSDirectory.open(getSearchIndexDirectory());
  }

  
  public QuestionPool searchQuestions(final String searchString) throws TrainerStorageException {
    try {
      IndexReader indexReader = DirectoryReader.open(getRAMIndexDir());
      IndexSearcher indexSearcher = new IndexSearcher(indexReader);
      
      Analyzer analyzer = new StandardAnalyzer(GermanAnalyzer.getDefaultStopSet());
      QueryParser queryParser = new QueryParser(SEARCH_INDEX_FIELD_CONTENTS, analyzer);
      queryParser.setDefaultOperator(QueryParser.Operator.AND);
      queryParser.setAllowLeadingWildcard(true);
      final Query query = queryParser.parse(searchString.replaceAll("[+\\-&|!\\(\\){}\\[\\]\\^\"~:\\/]", " "));
      final TopDocs hits = indexSearcher.search(query, Integer.MAX_VALUE);
      
      final ArrayList<Integer> questionIds = new ArrayList<Integer>();
      for (ScoreDoc scoreDoc : hits.scoreDocs) {
        Document document = indexSearcher.doc(scoreDoc.doc);
        questionIds.add(Integer.parseInt(document.get(SEARCH_INDEX_FIELD_ID)));
      }
      return allQuestions.getQuestionPoolForIds(questionIds);
    } catch (IOException e) {
      logger.error("Fehler beim Durchsuchen des Fragen-Indexes!", e);
      throw new TrainerStorageException();
    } catch (ParseException e) {
      logger.error("Fehler beim Durchsuchen des Fragen-Indexes!", e);
      throw new TrainerStorageException();
    }
  }
  
  private Path getSearchIndexDirectory() {
    return new File(KreawiGlobals.getApplicationDirectory(), APP_DATA_SEARCH_INDEX_DIR_NAME).toPath();
  }

  private void loadAllQuestions() {

    QuestionPool questions = null;
    
    
    try {
      if (!KreawiGlobals.isDemo()) {
        questions = loadQuestions(KreawiGlobals.QUESTION_DIR_NAME,
            KreawiGlobals.QUESTION_FILE_NAME);
      } else {
        File workDir = new File(System.getProperty("user.dir"));
        final File file = new File(new File(workDir, KreawiGlobals.DEMO_DIR_NAME), KreawiGlobals.QUESTION_DEMO_FILE_NAME);

        if (KreawiGlobals.verifyFileIntegrity(file, KreawiGlobals.DEMO_QUESTION_FILE_SIZE,
            KreawiGlobals.DEMO_QUESTION_FILE_HASH)) {
          
          questions = loadQuestions(KreawiGlobals.DEMO_DIR_NAME, KreawiGlobals.QUESTION_DEMO_FILE_NAME);
        } else {
          throw new QuestionReadWriteException(
              "Unable to load demo questions! (Question file verification failed!)");
        }
      }
    } catch (InvalidReaderWriterException e) {
      logger.error("Cannot read questions because of an invalid reader! (" + e.toString() + ") Trainer will be stopped.", e);
      logger.error("Cause: ", e.getCause());
      System.exit(1);
    } catch (QuestionReadWriteException e) {
      logger.error("Questions cannot be read! Trainer will be stopped.", e);
      logger.error("Cause: ", e.getCause());
      System.exit(1);
    } catch (NoSuchAlgorithmException e) {
      logger.error("Questions cannot be read! Trainer will be stopped.", e);
      logger.error("Cause: ", e.getCause());
      System.exit(1);
    } catch (IOException e) {
      logger.error("Questions cannot be read! Trainer will be stopped.", e);
      logger.error("Cause: ", e.getCause());
      System.exit(1);
    }
    
    if (!KreawiGlobals.isDemo()) {
      loadQuestionUpdates(questions);
    }

    allQuestions = questions;
  }

  private QuestionPool loadQuestions(String dir, String filename)
      throws InvalidReaderWriterException, QuestionReadWriteException {

    Properties properties = new Properties();
    properties.put(ReaderWriter.CONFIG_DIRECTORY_KEY, KreawiGlobals.getApplicationDirectory().getAbsolutePath());
    final File workDir = new File(System.getProperty("user.dir"));
    final File f = new File(new File(workDir, dir), filename);
    logger.info("Versuche Fragen aus Datei '" + f.getAbsolutePath() + "' zu lesen...");
    if (!f.exists()) {
      throw new QuestionReadWriteException("Datei '" + f.getAbsolutePath() + "' nicht gefunden!");
    }
    properties.put(ReaderWriter.DIRECTORY_KEY, f.getParent());
    properties.put(ReaderWriter.FILE_NAME_KEY, filename);

    // WORKAROUND: Scheinbar wurde mit einer älteren Version des Trainers ein anderer Reader angegeben
    //             (KreawiBinaryQuestionReader). Diese Konfigurationseinstellung wird dann nicht gelöscht?!
    //             Der Trainer startet nur ganz kurz und verschwindet dann ohne Fehlermeldung
    final String questionReaderClassName = "com.jquizengine.questions.readerwriter.BinaryQuestionReader";
//  String questionReaderClassName = this.model.getConfig().getString("questionReader");
    logger.info("Trying to use reader: " + questionReaderClassName);

    QuestionReader reader = new ReaderWriterFactory()
        .getReader(questionReaderClassName);

    reader.configure(properties);
    ProgressListener progressListener = new ProgressListener() {
      public void progressChange(ProgressChangeEvent e) {
        // do nothing
      }
    };
    reader.addProgressListener(progressListener);
    
    return reader.readQuestions();

  }

  private void loadQuestionUpdates(QuestionPool willBeUpdated) {
    
    try {
      QuestionPool updates = loadQuestions(KreawiGlobals.QUESTION_UPDATES_DIR_NAME,
          KreawiGlobals.QUESTION_UPDATES_FILE_NAME);
      
      Collection<Question> questionUpdates = updates.getQuestions();
      
      if (KreawiGlobals.isDevMode()) {
        String idsString = "";
        for (Question q : questionUpdates) {
          idsString += q.getId() + ",";
        }
        logger.info(idsString);
      }
      
      willBeUpdated.addAllQuestions(questionUpdates);
      
      logger.info("A total of " + questionUpdates.size() + " question updates were loaded.");
      
    } catch (InvalidReaderWriterException e) {
      logger.error("Question updates could not be loaded" + " due to faulty reader config!", e);
    } catch (QuestionReadWriteException e) {
      logger.info("No question updates were loaded. (" + e.toString() + ")");
    }
    
  }
  
  public QuestionPool getAllQuestions() {
    return this.allQuestions;
  }

  private Set<Integer> loadOldIntensiveCareUnit(final File sourceAppDir) throws TrainerStorageException {
    try {
      
      final File file = new File(sourceAppDir, getConfig().getString("intensiveCareFileName"));
      if (!file.exists()) {
        // initialize with empty ICU
        return new HashSet<Integer>();
      } else {
        return (Set<Integer>) TrainerUtil.readObjectFromFile(file);
      }
    } catch (ClassNotFoundException e) {
      logger.error("Fehler beim Laden der Intensivstation!", e);
      throw new TrainerStorageException();
    } catch (IOException e) {
      logger.error("Fehler beim Laden der Intensivstation!", e);
      throw new TrainerStorageException();
    }
  }
  
  @SuppressWarnings("unchecked")
  private Map<Integer, String> loadOldNotes(final File sourceAppDir) throws TrainerStorageException {
    try {
      final File file = new File(sourceAppDir, getConfig().getString("notesFileName"));
      if (!file.exists()) {
        // initialize with empty notes
        return new HashMap<Integer, String>(); 
      } else {
        return (Map<Integer, String>) TrainerUtil.readObjectFromFile(file);
      }
    } catch (ClassNotFoundException e) {
      logger.error("Fehler beim Laden der Notizen!", e);
      throw new TrainerStorageException();
    } catch (IOException e) {
      logger.error("Fehler beim Laden der Notizen!", e);
      throw new TrainerStorageException();
    }
  }
  
  private de.kreawi.trainer.QuestionAttributesModel loadOldQuestionAttributes(final File sourceAppDir)
      throws TrainerStorageException {
    try {
      final File file = new File(sourceAppDir, getConfig().getString("questionAttributesFileName"));
      if (!file.exists()) {
        // initialize with empty question attributes
        return new de.kreawi.trainer.QuestionAttributesModel();
      } else {
        return (de.kreawi.trainer.QuestionAttributesModel) TrainerUtil.readObjectFromFile(file);
      }
    } catch (ClassNotFoundException e) {
      logger.error("Fehler beim Laden des Karteikastens!", e);
      throw new TrainerStorageException();
    } catch (IOException e) {
      logger.error("Fehler beim Laden des Karteikastens!", e);
      throw new TrainerStorageException();
    }
  }
  
  public Properties loadOldSettingsFile(final File sourceAppDir) {
    final Properties settings = new Properties();
    final File file = new File(sourceAppDir, getConfig().getString("settingsFileName"));
    try (FileInputStream fis = new FileInputStream(file)) {
      settings.loadFromXML(fis);
    } catch (Exception e) {
      logger.error("Could not read old settings file (" + file.toString() + "). " +
                   "Using empty defaults.", 
                   e);
    }
    
    return settings;
  }
  
  public static Properties loadOldPropertiesDefaults() {
    final Properties defaults = new Properties();
    try (final InputStream in = TrainerUtil.class.getClassLoader().getResourceAsStream(INTERNAL_SETTINGS_FILE_PATH)) {
      defaults.loadFromXML(in);
    } catch (Exception e) {
      logger.error("Could not read defaults from internal file (" + INTERNAL_SETTINGS_FILE_PATH + "). " +
                   "Using empty defaults.", 
                   e);
    }
    return defaults;
  }

  public ConfigFile getConfig() {
    return getConfig(false);
  }

  private ConfigFile getConfig(final boolean noInit) {
    if (this.config == null && noInit) {
      return new ConfigFile(KreawiGlobals.CONFIG_FILE_NAME,
          KreawiGlobals.INTERNAL_CONFIG_FILE_PATH, KreawiGlobals.getApplicationDirectory(), false);
    } else if (this.config == null) {
      this.config = new ConfigFile(KreawiGlobals.CONFIG_FILE_NAME,
          KreawiGlobals.INTERNAL_CONFIG_FILE_PATH, KreawiGlobals.getApplicationDirectory(), !KreawiGlobals.isDemo());
    }
    return this.config;
  }

  public List<HistoryItem> getHistoryItemsMc() {
    return this.historyItemsMc;
  }

  public List<HistoryItem> getHistoryItemsText() {
    return this.historyItemsText;
  }
  
  private void loadHistoryItemsMc() throws TrainerStorageException {
    this.historyItemsMc = loadHistoryItems(HISTORY_MC_FILE_SUFFIX);
  }

  private void loadHistoryItemsText() throws TrainerStorageException {
    this.historyItemsText = loadHistoryItems(HISTORY_TEXT_FILE_SUFFIX);
  }
  
  private List<HistoryItem> loadHistoryItems(final String fileSuffix) throws TrainerStorageException {
    final ArrayList<HistoryItem> items = new ArrayList<HistoryItem>();
    final FilenameFilter filenameFilter = new FilenameFilter() {
      public boolean accept(File dir, String name) {
        return name.endsWith(fileSuffix);
      }
    };
      
    final File[] files = getHistoryDirectory().listFiles(filenameFilter);

    Arrays.sort(files, new Comparator<File>() {
      public int compare(File o1, File o2) {
        return (new Long(getStartTimestamp(o1)))
            .compareTo(new Long(getStartTimestamp(o2)))
            * -1;
      }
    });
    
    try {
      for (final File file : files) {
        try {
          final KreawiQuiz quiz = QuizModel.loadQuizFromFile(file, getAllQuestions());
          final HistoryItem item = createHistoryItem(file, quiz, items);
          items.add(item);
        } catch (IOException e) {
          logger.error("Error on loading history item file. Deleting file '" + file.toString() + "'");
          file.delete();
        }
      }
    } catch (ClassNotFoundException e) {
      logger.error("Fehler beim Laden des Verlaufs!", e);
      throw new TrainerStorageException();
    } catch (ClassCastException e) {
      logger.error("Fehler beim Laden des Verlaufs!", e);
      throw new TrainerStorageException();
    } catch (CloneNotSupportedException e) {
      logger.error("Fehler beim Laden des Verlaufs!", e);
      throw new TrainerStorageException();
    }
    return items;
  }

  private void addHistoryItem(final File file, final KreawiQuiz quiz) {
    final List<HistoryItem> items;
    if (quiz.containsOnlyTextQuestions()) {
      items = historyItemsText;
    } else {
      items = historyItemsMc;
    }
    items.add(0, createHistoryItem(file, quiz, items));
  }

  private HistoryItem createHistoryItem(final File file, final KreawiQuiz quiz, final List<HistoryItem> items) {
    final HistoryItem item = new HistoryItem(quiz);
    item.setDeleteAction(new Runnable() {
      public void run() {
        items.remove(item);
        file.delete();
      }
    });
    return item;
  }
  
  public void writeQuizToDisk(final KreawiQuiz quiz) throws TrainerStorageException {
    try {
      logger.info("Trying to write quiz to disk...");
      final File file = new File(getHistoryDirectory(), getFilenameForQuizHistory(quiz));
      final boolean updateExistingFile = file.exists();
      TrainerUtil.writeObjectToFile(quiz, file);
      logger.info("Success!");
      if (!updateExistingFile) {
        addHistoryItem(file, quiz);
      }
      if (quiz.containsOnlyTextQuestions()) {
        notifyTextHistoryChangedListeners();
      } else {
        notifyMcHistoryChangedListeners();
      }
    } catch (IOException e) {
      logger.error("Unable to write quiz to disk!", e);
      throw new TrainerStorageException();
    }
  }

  private String getFilenameForQuizHistory(final Quiz quiz) {
    return quiz.getStartTime()
        + (quiz.containsOnlyTextQuestions() ? HISTORY_TEXT_FILE_SUFFIX : HISTORY_MC_FILE_SUFFIX);
  }

  public static long getStartTimestamp(File file) {
    String timestamp = file.getName().substring(0, file.getName().indexOf(HISTORY_FILE_SUFFIX_SEPARATOR));
    return Long.parseLong(timestamp);
  }

  public File getHistoryDirectory() {
    File histDir = new File(KreawiGlobals.getApplicationDirectory(), APP_DATA_HISTORY_DIR_NAME);
    histDir.mkdirs();

    return histDir;
  }
  
  public void backupUserData(String backupTitle) throws TrainerStorageException {
    
    final long backupTime = System.currentTimeMillis();

    final File backupDir = getBackupDir(backupTime);

    final TrainerBackup backup = new TrainerBackup(backupTitle, backupTime, backupDir);
    
    createBackup(backup);
  }

  private void createBackup(final TrainerBackup backup) throws TrainerStorageException {
    try {
      final File backupDir = backup.getBackupDir();
      if (backupDir.exists()) {
        logger.error("Ausgabeverzeichnis für Backup schon vorhanden!");
        throw new TrainerStorageException();
      } else {
        backupDir.mkdirs();
      }
      
      final File backupFile = new File(backupDir, BACKUP_META_FILE_NAME);
      TrainerUtil.writeObjectToFile(backup, backupFile);
      
      final File dbBackupFile = new File(backupDir, BACKUP_DB_FILE_NAME);
      final String sql = createBackupSql(dbBackupFile.getAbsolutePath(),KreawiGlobals.DB_PASSWORD);
      final Connection connection = this.dataSourcePooled.getConnection();
      connection.createStatement().execute(sql);
      connection.close();
      
      FileUtils.copyDirectory(getHistoryDirectory(), new File(backupDir, APP_DATA_HISTORY_DIR_NAME));
      
    } catch (IOException | SQLException e) {
      logger.error("Fehler beim Erstellen der Sicherung!", e);
      throw new TrainerStorageException();
    }
  }

  private static String createBackupSql(final String dbBackupFile, final String password) {
    return "SCRIPT DROP TO '" + dbBackupFile + "' COMPRESSION LZF CIPHER AES PASSWORD '" + password + "'" + 
        " TABLE " + QuestionAttributes.getTableName();
  }
  
  public TrainerBackup[] readBackups() throws TrainerStorageException {
    try {
      final File allBackupsDir = new File(KreawiGlobals.getApplicationDirectory(), BACKUP_DIR_NAME);
      final String[] directories;
      final ArrayList<TrainerBackup> backups = new ArrayList<TrainerBackup>();
      if (allBackupsDir.exists()) {
        directories = allBackupsDir.list(new FilenameFilter() {
          public boolean accept(File current, String name) {
            return new File(current, name).isDirectory();
          }
        });
        if (directories.length > 0) {
          for (String dir : directories) {
            final File backupDir = new File(allBackupsDir, dir);
            final File file = new File(backupDir, BACKUP_META_FILE_NAME);
            if (file.exists()) {
              backups.add((TrainerBackup) TrainerUtil.readObjectFromFile(file));
            } else {
              logger.warn("Backup file '" + file.getAbsolutePath() + "' not found. Ignoring backup directory!");
            }
          }
          // sort by creation time in reverse order (newest first)
          Collections.sort(backups, new Comparator<TrainerBackup>() {
            public int compare(TrainerBackup b1, TrainerBackup b2) {
              return Long.valueOf(b1.getTime()).compareTo(Long.valueOf(b2.getTime())) * -1;
            }
          });
        }
      }
      
      List<TrainerBackup> olderVersionImports = readOlderVersionImports();
      backups.addAll(olderVersionImports);
      
      TrainerBackup[] backupsArray = new TrainerBackup[backups.size()];
      backupsArray = backups.toArray(backupsArray);
      
      return backupsArray;
    } catch (FileNotFoundException e) {
      logger.error("Fehler beim Lesen der Backups-Liste!", e);
      throw new TrainerStorageException();
    } catch (ClassNotFoundException e) {
      logger.error("Fehler beim Lesen der Backups-Liste!", e);
      throw new TrainerStorageException();
    } catch (IOException e) {
      logger.error("Fehler beim Lesen der Backups-Liste!", e);
      throw new TrainerStorageException();
    }
  }
  
  public boolean existsOlderVersionImport() {
    return readOlderVersionImports().size() > 0;
  }
  
  private List<TrainerBackup> readOlderVersionImports() {
    
    final ArrayList<TrainerBackup> olderVersionImports = new ArrayList<TrainerBackup>();
    
    for (int y=KreawiGlobals.MINIMUM_IMPORT_YEAR; y < KreawiGlobals.CURRENT_YEAR_INT; y++) {
      final File backupDir = new File(KreawiGlobals.getApplicationDirectory().getParent(),
          KreawiGlobals.APP_DATA_DIR_NAME_PREFIX + y);
      if (backupDir.exists() && backupDir.isDirectory()) {
        final String backupTitle = TrainerTextResources.getStringWithSubstitution(
            "RESTORE_USER_DATA_OLD_VERSION_IMPORT", Long.toString(y));
        final long time = backupDir.lastModified();
        final boolean isDbBackup = y >= 2019;  // DB backup since 2019
        final boolean otherTrainer = true;
        olderVersionImports.add(new TrainerBackup(backupTitle, time, backupDir, isDbBackup, otherTrainer));
      }
    }
    
    // newest import first
    Collections.reverse(olderVersionImports);
    
    return olderVersionImports;
  }
  
  public void restoreBackup(final TrainerBackup backup) throws TrainerStorageException {
    
    if (backup.isDbBackup()) {
      restoreBackupDb(backup);
    } else {
      // import settings from trainer 2018 or earlier
      restoreBackupOldBinFiles(backup.getBackupDir());
    }

    // reload history
    loadHistoryItemsMc();
    loadHistoryItemsText();
    
    notifyMcHistoryChangedListeners();
    notifyTextHistoryChangedListeners();

    try {
      this.settingsModel.notifyAllSettingsChangedListeners();
      this.questionAttributesModel.notifyCardBoxChangedListeners();
      this.questionAttributesModel.notifyIcuChangedListeners();
      this.questionAttributesModel.notifyNoteChangedListeners();
    } catch (Exception e) {
      logger.error("Fehler beim Wiederherstellen der Sicherung (bei Notifications)!", e);
      throw new TrainerStorageException();
    }

  }

  private void restoreBackupDb(final TrainerBackup backup) throws TrainerStorageException {
    try {
      // make backupDir relative to current settings directory
      // because absolute path is wrong if settings directory copied from other PC
      final File backupDir = new File(new File(KreawiGlobals.getApplicationDirectory(), BACKUP_DIR_NAME),
          backup.getBackupDir().getName());
      
      final File dbBackupFile;
      if (backup.isOtherTrainer()) {
        // create backup script for old db files
        final String dbUrl = KreawiGlobals.DB_PREFIX
            + new File(new File(backup.getBackupDir(), KreawiGlobals.DB_DIR_NAME), KreawiGlobals.DB_NAME)
                .getAbsolutePath();
        final Path tmpDir = Files.createTempDirectory("trainer_migration");
        dbBackupFile = new File(tmpDir.toFile(), BACKUP_DB_FILE_NAME);
        
        final Connection conn = DriverManager.getConnection(dbUrl, KreawiGlobals.DB_USER, KreawiGlobals.DB_PASSWORD);
        conn.createStatement().execute(createBackupSql(dbBackupFile.getAbsolutePath(), KreawiGlobals.DB_PASSWORD));
        conn.close();
      } else {
        dbBackupFile = new File(backupDir, BACKUP_DB_FILE_NAME);
      }

      // Restore backup script to temporary db file, link to real db and copy the tables
      final File tmpDir = Files.createTempDirectory("trainer").toFile();
      
      final String tmpDbUrl = KreawiGlobals.DB_PREFIX + (new File(tmpDir, TMP_DB_NAME)).getAbsolutePath(); 
      final Connection tmpDbConn = DriverManager.getConnection(tmpDbUrl, KreawiGlobals.DB_USER,
          KreawiGlobals.DB_PASSWORD);
      tmpDbConn.createStatement().execute("DROP ALL OBJECTS"); // make sure db is empty
      tmpDbConn.createStatement().execute("RUNSCRIPT FROM '" + dbBackupFile.getAbsolutePath()
          + "' COMPRESSION LZF CIPHER AES PASSWORD '" + KreawiGlobals.DB_PASSWORD + "'");
      tmpDbConn.close();
      
      final String tableNameOriginal = QuestionAttributes.getTableName();
      final String tableNameRestore = tableNameOriginal + "_restore";
      final Connection conn = this.dataSourcePooled.getConnection();
      conn.createStatement().execute("DELETE FROM " + tableNameOriginal + " WHERE True");
      conn.createStatement().execute("DROP TABLE IF EXISTS " + tableNameRestore);
      conn.createStatement().execute("CREATE LINKED TABLE " + tableNameRestore + "('', '" + tmpDbUrl + "', '"
          + KreawiGlobals.DB_USER + "', '" + KreawiGlobals.DB_PASSWORD + "', '" + tableNameOriginal + "')");
      // WARNING: can't use CREATE TABLE ... AS (SELECT ...) here because linked table looses default values
      conn.createStatement().execute("INSERT INTO " + tableNameOriginal + " SELECT * FROM " + tableNameRestore + "");
      conn.createStatement().execute("DROP TABLE " + tableNameRestore);
      conn.close();
      
      DeleteDbFiles.execute(tmpDir.getAbsolutePath(), TMP_DB_NAME, false);
      tmpDir.deleteOnExit();
      
      // Delete backup script if previously created from another trainer db
      if (backup.isOtherTrainer()) {
        dbBackupFile.delete();
        dbBackupFile.getParentFile().delete();
      }
      
      FileUtils.deleteDirectory(getHistoryDirectory());
      final File historyBackupDir = new File(backupDir, APP_DATA_HISTORY_DIR_NAME);
      FileUtils.copyDirectory(historyBackupDir, getHistoryDirectory());
    } catch (IOException | SQLException e) {
      logger.error("Fehler beim Wiederherstellen der Sicherung eines Trainers mit DB!", e);
      throw new TrainerStorageException();
    }
    
  }

  private void restoreBackupOldBinFiles(final File backupDir) throws TrainerStorageException {
    
    try {
      
      // init models to init DB tables!
      final QuestionAttributesModel tempQAaModel = new QuestionAttributesModel(this.dataSourcePooled);
      final TrainerSettingsModel tempSettingsModel = new TrainerSettingsModel(this.dataSourcePooled);
      
      final Set<Integer> icu = loadOldIntensiveCareUnit(backupDir);
      for (Integer questionId : icu) {
        tempQAaModel.setIcu(questionId, true);
      }
      
      final Map<Integer, String> notes = loadOldNotes(backupDir);
      for (Integer questionId : notes.keySet()) {
        tempQAaModel.setNote(questionId, notes.get(questionId));
      }
      
      final de.kreawi.trainer.QuestionAttributesModel oldQAs = loadOldQuestionAttributes(backupDir);
      for (Integer questionId : oldQAs.getQuestionIds()) {
        final int level = oldQAs.getCardBoxLevel(questionId);
        final long timestamp = oldQAs.getCardBoxTimestamp(questionId);
        tempQAaModel.setCardBox_DontUseThis(questionId, level, timestamp);
      }
      
      final Properties oldSettings = loadOldSettingsFile(backupDir);
      for (Object key_obj : oldSettings.keySet()) {
        if (key_obj instanceof String) {
          final String key = (String) key_obj;
          tempSettingsModel.setValue(key, oldSettings.getProperty(key));
        }
      }
      
      FileUtils.deleteDirectory(getHistoryDirectory());
      FileUtils.copyDirectory(new File(backupDir, APP_DATA_HISTORY_DIR_NAME), getHistoryDirectory());
    } catch (IOException e) {
      logger.error("Fehler beim Wiederherstellen der Sicherung eines Trainers ohne DB!", e);
      throw new TrainerStorageException();
    }
  }

  private File getBackupDir(final long backupTime) {
    final File backupDir = new File(new File(KreawiGlobals.getApplicationDirectory(), BACKUP_DIR_NAME),
        Long.toString(backupTime));
    
    return backupDir;
  }

  public OnlineGamePersistence getOnlineGamePersistence() {
    return onlineGamePersistence;
  }

  public QuestionAttributesModel getQuestionAttributesModel() {
    return questionAttributesModel;
  }

  public TrainerSettingsModel getSettingsModel() {
    return settingsModel;
  }
  
  public void addMcHistoryChangedListener(final McHistoryChangedListener listener) {
    this.mcHistoryChangedListeners .add(listener);
  }
  
  public void removeMcHistoryChangedListener(final McHistoryChangedListener listener) {
    this.mcHistoryChangedListeners.remove(listener);
  }
  
  private void notifyMcHistoryChangedListeners() {
    for (McHistoryChangedListener listener : this.mcHistoryChangedListeners) {
      listener.historyChanged();
    }
  }
  
  public interface McHistoryChangedListener {
    public void historyChanged();
  }
  
  public void addTextHistoryChangedListener(final TextHistoryChangedListener listener) {
    this.textHistoryChangedListeners.add(listener);
  }
  
  public void removeTextHistoryChangedListener(final TextHistoryChangedListener listener) {
    this.textHistoryChangedListeners.remove(listener);
  }
  
  private void notifyTextHistoryChangedListeners() {
    for (TextHistoryChangedListener listener : this.textHistoryChangedListeners) {
      listener.historyChanged();
    }
  }
  
  public interface TextHistoryChangedListener {
    public void historyChanged();
  }
  
}
