package de.kreawi.trainer.db;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import javax.sql.DataSource;

import org.javalite.activejdbc.Base;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class QuestionAttributesModel extends DbAccessModel {
  
  public final static int CARD_BOX_LEVEL_UNDEFINED = -1;
  
  private List<CardBoxChangedListener> cardBoxChangedListeners = new ArrayList<CardBoxChangedListener>();
  private List<IcuChangedListener> icuChangedListeners = new ArrayList<IcuChangedListener>();
  private List<NoteChangedListener> notesChangedListeners = new ArrayList<NoteChangedListener>();

  @SuppressWarnings("unused")
  private static final Logger logger = LoggerFactory.getLogger(QuestionAttributesModel.class); 

  public QuestionAttributesModel(final DataSource dataSource) {
    super(dataSource);
    runWithDbConn(() -> {
      QuestionAttributes.createTableIfNotExists();
    });
  }
  
  public void setCardBoxLevel(final int questionId, final int cardBoxLevel) {
    runWithDbConn(() -> {
      final QuestionAttributes qa = QuestionAttributes.findOrInit(QuestionAttributes.DB_QUESTION_ID, questionId);
      
      if (cardBoxLevel >= 0 || cardBoxLevel == CARD_BOX_LEVEL_UNDEFINED) {
        qa.setCardBoxLevel(cardBoxLevel);
        qa.setCardBoxTimestamp(System.currentTimeMillis());
      } else {
        throw new IllegalArgumentException("Card box level must be >=0 or CARD_BOX_LEVEL_UNDEFINED!");
      }
      
      qa.saveIt();
      
      notifyCardBoxChangedListeners();
    });
  }

  /** Only used for restoring backup from older (<= 2018) trainer
   *  No notification of listeners!  **/
  public void setCardBox_DontUseThis(final int questionId, final int level, final long timestamp) {
    runWithDbConn(() -> {
      final QuestionAttributes qa = QuestionAttributes.findOrInit(QuestionAttributes.DB_QUESTION_ID, questionId);
      qa.setCardBoxLevel(level);
      qa.setCardBoxTimestamp(timestamp);
      qa.saveIt();
    });
  }

  public long getCardBoxTimestamp(final int questionId) {
    return callWithDbConn(() -> {
      final QuestionAttributes qa = QuestionAttributes.findOrCreateIt(QuestionAttributes.DB_QUESTION_ID, questionId);
      return qa.getCardBoxTimestamp();
    });
  }
  
  public int getCardBoxLevel(int questionId) {
    return callWithDbConn(() -> {
      QuestionAttributes qa = QuestionAttributes.findOrCreateIt(QuestionAttributes.DB_QUESTION_ID, questionId);
      return qa.getCardBoxLevel();
    });
  }
  
  public Collection<Integer> getQuestionIdsForCardBoxLevel(final int level) {
    return callWithDbConn(() -> {
      final List<QuestionAttributes> qas = QuestionAttributes.where(QuestionAttributes.DB_CARD_BOX_LEVEL + "=?", level);
      return Collections.unmodifiableCollection(getIdsList(qas));
    });
  }

  public Collection<Integer> getQuestionIdsWithDefinedCardBoxLevel() {
    return callWithDbConn(() -> {
      final List<QuestionAttributes> qas = QuestionAttributes.where("NOT " + QuestionAttributes.DB_CARD_BOX_LEVEL + "=?",
        CARD_BOX_LEVEL_UNDEFINED);
      return Collections.unmodifiableCollection(getIdsList(qas));
    });
  }
  
  public Long getQuestionCountForCardBoxLevel(final int level) {
    return callWithDbConn(() -> {
      return QuestionAttributes.count(QuestionAttributes.DB_CARD_BOX_LEVEL + "=?", level);
    });
  }
  
  public Set<Integer> getIntensiveCareUnit() {
    return callWithDbConn(() -> {
      final List<QuestionAttributes> qas = QuestionAttributes.where(QuestionAttributes.DB_ICU + "=?", true);
      return Collections.unmodifiableSet(getIdsSet(qas));
    });
  }
  
  public void setIcu(final int questionId, final boolean icu) {
    runWithDbConn(() -> {
      final QuestionAttributes qa = QuestionAttributes.findOrInit(QuestionAttributes.DB_QUESTION_ID, questionId);
      qa.setIcu(icu);
      qa.saveIt();
      
      notifyIcuChangedListeners();
    });
  }
  
  public boolean isInIcu(final int questionId) {
    return callWithDbConn(() -> {
      final QuestionAttributes qa = QuestionAttributes.findOrInit(QuestionAttributes.DB_QUESTION_ID, questionId);
      return qa.isIcu();
    });
  }
  
  public Map<Integer, String> getNotes() {
    return callWithDbConn(() -> {
      final List<QuestionAttributes> qas = QuestionAttributes
          .where("NOT " + QuestionAttributes.DB_NOTE + " IS NULL AND NOT " + QuestionAttributes.DB_NOTE + "=''", true);
      final Map<Integer, String> map = qas.stream()
          .collect(Collectors.toMap(QuestionAttributes::getQuestionId, QuestionAttributes::getNote));
      return Collections.unmodifiableMap(map);
    });
  }
  
  public String getNote(final int questionId) {
    return callWithDbConn(() -> {
      final QuestionAttributes qa = QuestionAttributes.findOrInit(QuestionAttributes.DB_QUESTION_ID, questionId);
      return qa.getNote();
    });
  }

  public void setNote(final int questionId, final String note) {
    runWithDbConn(() -> {
      final QuestionAttributes qa = QuestionAttributes.findOrInit(QuestionAttributes.DB_QUESTION_ID, questionId);
      qa.setNote(note);
      qa.saveIt();
      
      notifyNoteChangedListeners();
    });
  }
  
  private List<Integer> getIdsList(final List<QuestionAttributes> qas) {
    return qas.stream().map(qa -> new Integer(qa.getQuestionId())).collect(Collectors.toList());
  }
  
  private Set<Integer> getIdsSet(final List<QuestionAttributes> qas) {
    return qas.stream().map(qa -> new Integer(qa.getQuestionId())).collect(Collectors.toSet());
  }

  public void resetCardBox() {
    runWithDbConn(() -> {
      QuestionAttributes.update(
          QuestionAttributes.DB_CARD_BOX_LEVEL + "=?, " + QuestionAttributes.DB_CARD_BOX_TIMESTAMP + "=?", 
          "True",
          CARD_BOX_LEVEL_UNDEFINED, 0);
      
      notifyCardBoxChangedListeners();
    });
  }
  
  public void resetIcu() {
    runWithDbConn(() -> {
      QuestionAttributes.update(QuestionAttributes.DB_ICU + "=?", "True", false);
      
      notifyIcuChangedListeners();
    });
  }

  public void resetNotes() {
    runWithDbConn(() -> {
      Base.exec(
          "UPDATE " + QuestionAttributes.getTableName() + " SET " + QuestionAttributes.DB_NOTE + " = NULL WHERE True");
      
      notifyNoteChangedListeners();
    });
  }

  public void addCardBoxChangedListener(final CardBoxChangedListener listener) {
    this.cardBoxChangedListeners.add(listener);
  }

  public void notifyCardBoxChangedListeners() {
    for (CardBoxChangedListener listener : this.cardBoxChangedListeners) {
      listener.cardBoxLevelChanged();
    }
  }

  public interface CardBoxChangedListener {
    public void cardBoxLevelChanged();
  }

  public void addIcuChangedListener(final IcuChangedListener listener) {
    this.icuChangedListeners.add(listener);
  }

  public void notifyIcuChangedListeners() {
    for (IcuChangedListener listener : this.icuChangedListeners) {
      listener.icuChanged();
    }
  }

  public interface IcuChangedListener {
    public void icuChanged();
  }

  public void addNoteChangedListener(final NoteChangedListener listener) {
    this.notesChangedListeners.add(listener);
  }

  public void notifyNoteChangedListeners() {
    for (NoteChangedListener listener : this.notesChangedListeners) {
      listener.noteChanged();
    }
  }

  public interface NoteChangedListener {
    public void noteChanged();
  }

}
