package de.kreawi.trainer.db;

import java.util.List;
import java.util.stream.Collectors;

import javax.sql.DataSource;

import org.javalite.activejdbc.validation.ValidationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.onlinegame.OnlineGamePersistence;
import de.kreawi.trainer.onlinegame.OnlineGameSettings;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuiz;
import de.kreawi.trainer.onlinegame.entities.OnlineGameUser;

public class TrainerOnlineGamePersistence extends DbAccessModel implements OnlineGamePersistence {
  
  private final OnlineGameSettings settings; 

  private final Logger logger = LoggerFactory.getLogger(TrainerOnlineGamePersistence.class);

  public TrainerOnlineGamePersistence(final DataSource dataSource) {
    super(dataSource);
    runWithDbConn(() -> {
      OnlineGameUser.createTableIfNotExists();
      OnlineGameQuiz.createTableIfNotExists();
    });
    
    this.settings = new TrainerOnlineGameSettings(this, dataSource);
    
  }

  public OnlineGameSettings getSettings() {
    return this.settings;
  }

  /* -------------------------- OnlineGameQuiz -------------------------------------- */

  @Override
  public void addNewQuiz(OnlineGameQuiz quiz) {
    
    insertQuiz(quiz);
    
    // preload opponent
    runWithDbConn(() -> {
      quiz.getOpponent();
    });
    
  }


  private void insertQuiz(final OnlineGameQuiz quiz) {
    runWithDbConn(() -> {
      if (!quiz.insert()) {
        final String message = "Quiz konnte nicht angelegt werden!" + 
            (quiz.hasErrors()?"(Fehler: " + quiz.errors().toString() + ")":"");
        logger.error(message);
      }
    });
  }

  @Override
  public void updateQuiz(final OnlineGameQuiz quiz) {
    runWithDbConn(() -> {
      try {
        quiz.saveIt();
      } catch (ValidationException e) {
        final String message = "Quiz konnte nicht gespeichert werden!";
        logger.error(message, e);
      }
    });
  }

  @Override
  public void removeQuiz(OnlineGameQuiz quiz) {
    runWithDbConn(() -> {
      quiz.delete();
    });
  }
  
  /* (non-Javadoc)
   * @see de.kreawi.trainer.db.OnlineGamePersistence#getCurrentQuizzes()
   */
  @SuppressWarnings("unchecked")
  @Override
  public List<OnlineGameQuiz> getCurrentQuizzes() {
    return callWithDbConn(() -> {
      return OnlineGameQuiz.where("status NOT IN (" + getFinishedStatus() + ")")
          .orderBy(OnlineGameQuiz.DB_CREATED_AT + " DESC").include(OnlineGameUser.class).load();
    });
  }

  /* (non-Javadoc)
   * @see de.kreawi.trainer.db.OnlineGamePersistence#getFinishedQuizzes()
   */
  @SuppressWarnings("unchecked")
  @Override
  public List<OnlineGameQuiz> getFinishedQuizzes() {
    return callWithDbConn(() -> {
      return OnlineGameQuiz.where("status IN (" + getFinishedStatus() + ")")
      .orderBy(OnlineGameQuiz.DB_CREATED_AT + " DESC").include(OnlineGameUser.class).load();
    });
  }

  private String getFinishedStatus() {
    final StringBuffer b = new StringBuffer();
    b.append("'");
    b.append(OnlineGameQuiz.getFinishedStatus().stream().map(Object::toString).collect(Collectors.joining("', '")));
    b.append("'");
    return b.toString();
  }


  /* -------------------------- End OnlineGameQuiz -------------------------------------- */

  
  /* -------------------------- OnlineGameUser -------------------------------------- */
  @Override
  public OnlineGameUser getUser(final String id) {
    return callWithDbConn(() -> {
      return OnlineGameUser.findById(id);
    });
  }

  private void insertUser(final OnlineGameUser user) {
    runWithDbConn(() -> {
      user.insert();
    });
  }

  private void updateUser(final OnlineGameUser user) {
    runWithDbConn(() -> {
      user.saveIt();
    });
    
  }

  public void updateOrInsertUser(OnlineGameUser user) {
    // id is delivered by the server, so activejdbc's save method cannot determine if record already exists
    
    // check if already in DB
    final boolean newUser = callWithDbConn(() -> {
      final OnlineGameUser u = OnlineGameUser.findById(user.getId());
      return (u == null);
    });
    
    if (newUser) {
      insertUser(user);
    } else {
      updateUser(user);
    }
  }

  @Override
  public void resetAllData() {
    runWithDbConn(() -> {
      TrainerOnlineGameSetting.deleteAll();
      OnlineGameQuiz.deleteAll();
      OnlineGameUser.deleteAll();
    });
  }
  
}
