package de.kreawi.trainer.dialogs.onlineGame.custom_nodes;

import java.util.ArrayList;
import java.util.Collection;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.dialogs.onlineGame.OnlineGameContainer;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuiz;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuiz.QuizStatus;
import de.kreawi.trainer.onlinegame.entities.StatusListener;
import javafx.application.Platform;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.control.Button;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Region;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.shape.Circle;
import javafx.scene.text.Font;
import javafx.scene.text.FontWeight;
import javafx.scene.text.Text;

public class QuizListItemView extends BorderPane {

  private final OnlineGameQuiz quiz;

  private final OnlineGameContainer ogc;

  @SuppressWarnings("unused")
  private static final Logger logger = LoggerFactory.getLogger(QuizListItemView.class); 

  public QuizListItemView(final OnlineGameQuiz quiz, final OnlineGameContainer ogc) {
    
    this.quiz = quiz;
    this.ogc = ogc;
    
    this.quiz.addStatusListener(new StatusListener() {
      @Override
      public void statusChanged(OnlineGameQuiz quiz, QuizStatus oldStatus, QuizStatus newStatus) {
        logger.info("Received status change. Reinit infos.");
        Platform.runLater(() -> {
          init();
        });
      }
    });
    
    init();
    
  }

  private void init() {
    
    setMinHeight(70.0);
    
    final Circle dot = new Circle(5, Color.web("#55c"));
    final boolean showDot = 
        QuizStatus.INVITATION.equals(this.quiz.getStatus()) ||
        QuizStatus.ACCEPTED.equals(this.quiz.getStatus()) ||
        QuizStatus.SELECT_TOPIC.equals(this.quiz.getStatus()) ||
        QuizStatus.ROUND_PLAYING.equals(this.quiz.getStatus());
    dot.setVisible(showDot);
    
    final HBox imageDot = new HBox(5, dot, new ProfileImage(this.quiz.getOpponent()));
    imageDot.setPadding(new Insets(0, 5, 0, 5));
    imageDot.setAlignment(Pos.CENTER);
    
    setLeft(imageDot);
    
    final BorderPane captionAndButtons = new BorderPane();
    captionAndButtons.minWidthProperty().bind(minWidthProperty().subtract(imageDot.widthProperty()));
    captionAndButtons.minHeightProperty().bind(minHeightProperty());
    
    final Collection<Button> statusButtons = generateButtons();
    final Region buttons;
    if (!statusButtons.isEmpty()) {
      final HBox hbox = new HBox(10d);
      hbox.getChildren().addAll(statusButtons);
      hbox.setPadding(new Insets(0, 5, 0, 5));
      hbox.minHeightProperty().bind(minHeightProperty());
      hbox.setAlignment(Pos.CENTER);
      buttons = hbox; 
      captionAndButtons.setRight(buttons);
    } else {
      buttons = new Region();
    }
    
    final VBox captions = new VBox();
    captions.minWidthProperty().bind(captionAndButtons.minWidthProperty().subtract(buttons.widthProperty()));
    captions.setAlignment(Pos.CENTER_LEFT);

    final Text username = new Text(this.quiz.getOpponent().getUserName());
    username.setFont(Font.font("Arial", FontWeight.BOLD, 13));
    captions.getChildren().add(username);
    
    
    final String statusText = generateStatusText(this.quiz);
    if (statusText != null && !statusText.isEmpty()) {
      final Text status = new Text(statusText);
      status.setFont(Font.font("Arial", FontWeight.NORMAL, 11));
      status.setFill(Color.web("#555"));
      status.wrappingWidthProperty().bind(captions.minWidthProperty());
      captions.getChildren().add(status);
    }
    
    captionAndButtons.setCenter(captions);
    
//    captionAndButtons.setStyle("-fx-border-width: 1px; -fx-border-style: solid; -fx-border-color: green");
//    captions.setStyle("-fx-border-width: 1px; -fx-border-style: solid; -fx-border-color: orange");
//    buttons.setStyle("-fx-border-width: 1px; -fx-border-style: solid; -fx-border-color: red");
    
    
    setCenter(captionAndButtons);

  }

  private String generateStatusText(OnlineGameQuiz item) {
    switch (item.getStatus()) {
    case ACCEPTED:
      return "Sie haben die Einladung angenommen.";
    case INVITATION:
      return "Sie wurden von " + item.getOpponent().getUserName() + " eingeladen.";
    case INVITATION_SENT:
      return "Sie haben " + item.getOpponent().getUserName() + " eingeladen.";
    case PENDING:
      return item.getOpponent().getUserName() + " hat Ihre Einladung angenommen.";
    case REJECTED:
      return "Die Einladung wurde abgelehnt.";
    case ROUND_PLAYING:
      return item.getOpponent().getUserName() + " wartet auf Ihr Spiel.";
    case ROUND_WAITING:
      return item.getOpponent().getUserName() + " spielt gerade.";
    case SELECT_TOPIC:
      return "Wählen Sie ein Thema für die nächste Runde!";
    case STARTED:
      return "Das Spiel wurde gestartet.";
    case WON:
      return "Sie haben gewonnen.";
    case LOST:
      return "Sie haben verloren.";
    case DRAW:
      return "Unentschieden.";
    case FINISHED:
      return "Das Spiel ist beendet.";
    case SUBMITTING_DATA:
      return "Warte auf Antwort vom Server.";
    default:
      break;
    
    }
    return null;
  }

  private Collection<Button> generateButtons() {
    final ArrayList<Button> buttons = new ArrayList<Button>();
    switch (this.quiz.getStatus()) {
    case INVITATION:
      final Runnable acceptAction = () -> {
        this.ogc.acceptInvitation(this.quiz);
      };
      final Runnable rejectAction = () -> {
        this.ogc.rejectInvitation(this.quiz);
      };
      buttons.add(new OGButton("Ja", "Einladung annehmen", acceptAction));
      buttons.add(OGButton.generateRejectButton("Einladung ablehnen", rejectAction));
      break;
    case ACCEPTED:
      final Runnable startAction = () -> {
        this.ogc.startQuiz(this.quiz);
      };
      buttons.add(new OGButton("Start", "Spiel starten", startAction));
      break;
    case ROUND_PLAYING:
      final Runnable playRoundAction = () -> {
        this.ogc.playRound(this.quiz);
      };
      buttons.add(new OGButton("Spielen", "Nächste Runde spielen", playRoundAction));
      break;
    case ROUND_WAITING:
      final Runnable showResultsAction = () -> {
        this.ogc.showQuizResults(quiz);
      };
      buttons.add(new OGButton("Spielstand", "Ergebnisse der letzten Runde(n) anzeigen.", showResultsAction));
      break;
    case SELECT_TOPIC:
      final Runnable selectTopicAction = () -> {
        this.ogc.selectTopic(this.quiz);
      };
      buttons.add(new OGButton("Spielen", "Thema für die nächste Runde wählen", selectTopicAction));
      break;
    case WON:
    case DRAW:
    case FINISHED:
    case LOST:
      final Runnable showFinalResultsAction = () -> {
        this.ogc.showQuizResults(quiz);
      };
      buttons.add(new OGButton("Ergebnisse", "Ergebnisse des Spiels anzeigen.", showFinalResultsAction));
      break;
    case INVITATION_SENT:
    case REJECTED:
    case PENDING:
    case STARTED:
    case SUBMITTING_DATA:
    default:
      break;
    
    }
    return buttons;
  }

}
