package de.kreawi.trainer.onlinegame;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.onlinegame.entities.OnlineGameNotification;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuiz;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuiz.QuizStatus;
import de.kreawi.trainer.onlinegame.entities.OnlineGameUser;
import de.kreawi.trainer.onlinegame.entities.StatusListener;

public class OnlineGameModel {
  
  public enum OnlineGameStatus {
    OFFLINE, 
    BUSY, 
    ONLINE
  }

  private OnlineGameStatus status = OnlineGameStatus.OFFLINE;
  
  private final OnlineGamePersistence persistence;

  private final ArrayList<QuizListChangedListener> quizListChangedListeners = new ArrayList<QuizListChangedListener>();
  
  // CopyOnWriteArrayList avoids ConcurrentModificationException if iterating list while being changed
  private CopyOnWriteArrayList<OnlineGameQuiz> currentQuizzes;
  private CopyOnWriteArrayList<OnlineGameQuiz> finishedQuizzes;
  
  private static final Logger logger = LoggerFactory.getLogger(OnlineGameModel.class); 

  public OnlineGameModel(final OnlineGamePersistence persistence) {
    this.persistence = persistence;

  }
  
  public void addMoveToFinishedListener(final OnlineGameQuiz quiz) {
    final StatusListener listener = new StatusListener() {
      @Override
      public void statusChanged(OnlineGameQuiz quiz, QuizStatus oldStatus, QuizStatus newStatus) {
        if (OnlineGameQuiz.getFinishedStatus().contains(newStatus)) {
          
          logger.info("Moving quiz to finished quizzes list...");
          
          getCurrentQuizzes().remove(quiz);
          getFinishedQuizzes().add(quiz);
          
          OnlineGameModel.this.notifyQuizListChangedListeners(quiz);
          
          quiz.removeStatusListener(this);
        }
      }
    };
    quiz.addStatusListener(listener);
  }

  public OnlineGameStatus getStatus() {
    return status;
  }
  
  public void setStatus(final OnlineGameStatus status) {
    this.status = status;
  }
  
  public void resetAllData() {
    this.persistence.resetAllData();
    
  }
  
  public OnlineGameSettings getSettings() {
    return this.persistence.getSettings();
  }
  
  public List<OnlineGameQuiz> getCurrentQuizzes() {
    if (this.currentQuizzes == null) {
      this.currentQuizzes = new CopyOnWriteArrayList<OnlineGameQuiz>();
      this.currentQuizzes.addAll(this.persistence.getCurrentQuizzes());
      this.currentQuizzes.forEach((quiz) -> {addMoveToFinishedListener(quiz);});

    }
    return this.currentQuizzes;
  }
  
  public List<OnlineGameQuiz> getFinishedQuizzes() {
    if (this.finishedQuizzes == null) {
      this.finishedQuizzes = new CopyOnWriteArrayList<OnlineGameQuiz>();
      this.finishedQuizzes.addAll(this.persistence.getFinishedQuizzes());
    }
    return this.finishedQuizzes;
  }
  
  public void addNewQuiz(final OnlineGameNotification notification) {

    final String quizId = notification.getQuizChallengeId();
    final OnlineGameUser opponent = createOpponentFromNotification(notification);
    final QuizStatus status = OnlineGameQuiz.getStatusFromNotificationType(notification.getType());
    final OnlineGameQuiz quiz = new OnlineGameQuiz(quizId, opponent, status, notification.getNotificationId());
    this.persistence.addNewQuiz(quiz);
    
    if (OnlineGameQuiz.getFinishedStatus().contains(quiz.getStatus())) {
      getFinishedQuizzes().add(quiz);
    } else {
      getCurrentQuizzes().add(quiz);
      addMoveToFinishedListener(quiz);
    }
    notifyQuizListChangedListeners(quiz);
  }
  
  public void removeQuiz(final OnlineGameQuiz quiz) {
    this.persistence.removeQuiz(quiz);
    
    getCurrentQuizzes().remove(quiz);
    getFinishedQuizzes().remove(quiz);
    
    notifyQuizListChangedListeners(null);
  }
  
  private OnlineGameUser createOpponentFromNotification(final OnlineGameNotification notification) {
    
    final String opponentId = notification.getSenderId();
    
    OnlineGameUser opponent = this.persistence.getUser(opponentId);
    
    final boolean isNewOpponent = opponent == null;
    
    if (isNewOpponent) {
      opponent = new OnlineGameUser();
      opponent.setId(opponentId);
    }
    
    opponent.setUserName(notification.getUserName());
    opponent.setProfileImage(notification.getProfileImage());
    opponent.setFriend(true);
    
    this.persistence.updateOrInsertUser(opponent);
    
    return opponent;
    
  }
  
  public void updateQuiz(final OnlineGameQuiz quiz) {
    this.persistence.updateQuiz(quiz);
  }

  public OnlineGameQuiz getCurrentQuiz(String quizChallengeId) {
    if (quizChallengeId == null) {
      return null;
    }
    for (OnlineGameQuiz quiz : this.getCurrentQuizzes()) {
      if (quizChallengeId.equals(quiz.getId())) {
        return quiz;
      }
    }
    return null;
  }
  
  public OnlineGameQuiz getFinishedQuiz(final String quizChallengeId) {
    if (quizChallengeId == null) {
      return null;
    }
    for (OnlineGameQuiz quiz : this.getFinishedQuizzes()) {
      if (quizChallengeId.equals(quiz.getId())) {
        return quiz;
      }
    }
    return null;
  }
  
  public String getServerSession() {
    return this.persistence.getSettings().getServerSession();
  }

  public void setServerSession(final String session) {
    this.persistence.getSettings().setServerSession(session);
  }

  public boolean hasServerSession() {
    return getServerSession() != null;
  }
  
  public void removeServerSession() {
    this.persistence.getSettings().setServerSession(null);
  }
  
  public void addQuizListChangedListener(final QuizListChangedListener listener) {
    this.quizListChangedListeners.add(listener);
  }
  
  public void removeQuizListChangedListener(final QuizListChangedListener listener) {
    this.quizListChangedListeners.remove(listener);
  }
  
  public void notifyQuizListChangedListeners(final OnlineGameQuiz newQuiz) {
    for (QuizListChangedListener listener : this.quizListChangedListeners) {
      listener.listChanged(newQuiz);
    }
  }
  
  public static interface QuizListChangedListener {
    public void listChanged(final OnlineGameQuiz newQuiz);
  }

}
