package de.kreawi.trainer.onlinegame.entities;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.javalite.activejdbc.Base;
import org.javalite.activejdbc.Model;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.onlinegame.entities.OnlineGameNotification.NotificationType;

public class OnlineGameQuiz extends Model {

  
  public static final String DB_ID = "id";
  public static final String DB_STATUS = "status";
  public static final String DB_OPPONENT_ID = "online_game_user_id";
  public static final String DB_LAST_NOTIFICATION_ID = "last_notification_id";
  public static final String DB_CREATED_AT = "created_at";
  public static final String DB_UPDATED_AT = "updated_at";

  private static Logger logger = LoggerFactory.getLogger(OnlineGameQuiz.class); 
  
  public static enum QuizStatus {
    INVITATION, 
    INVITATION_SENT, 
    ACCEPTED, 
    REJECTED, 
    PENDING, 
    SELECT_TOPIC, 
    STARTED, 
    ROUND_PLAYING, 
    ROUND_WAITING, 
    WON, 
    LOST, 
    DRAW, 
    FINISHED,
    SUBMITTING_DATA
  }
  
  public static List<QuizStatus> getFinishedStatus() {
    final QuizStatus[] array = new QuizStatus[] { QuizStatus.FINISHED, QuizStatus.WON, QuizStatus.DRAW, QuizStatus.LOST,
        QuizStatus.REJECTED };
    return Arrays.asList(array);
  }
  
  public boolean isFinished() {
    return getFinishedStatus().contains(getStatus());
  }

  private ArrayList<StatusListener> statusListeners = new ArrayList<StatusListener>();
  private ArrayList<StatusListener> addedStatusListeners = new ArrayList<StatusListener>();
  private ArrayList<StatusListener> removedStatusListeners = new ArrayList<StatusListener>();

  private OnlineGameResults results;
  private OnlineGameUser opponent;

  public OnlineGameQuiz() {
    // default constructor for ActiveJDBC
  }
  
  public OnlineGameQuiz(final String quizId, final OnlineGameUser opponent, final QuizStatus status, final String lastNotificationId) {
    
    setId(quizId);
    setOpponent(opponent);
    setStatus(status);
    setLastNotificationId(lastNotificationId);
    
  }

  /** Don't use any functions of activejdbc.Model before creating the table! **/ 
  public static void createTableIfNotExists() {
    final String createQuery = "CREATE TABLE IF NOT EXISTS online_game_quizzes (" +
        DB_ID + " VARCHAR(50) PRIMARY KEY, " +  
        DB_STATUS + " VARCHAR(30) NULL, " +  
        DB_OPPONENT_ID + " VARCHAR(50) NOT NULL, " +
        DB_LAST_NOTIFICATION_ID + " VARCHAR(50) NULL, " +
        DB_CREATED_AT + " TIMESTAMP NULL, " + 
        DB_UPDATED_AT + " TIMESTAMP NULL " + 
        ");";
    Base.exec(createQuery);
  }

  public String getId() {
    return getString(DB_ID);
  }

  public void setId(final String id) {
    setString(DB_ID, id);
  }

  public OnlineGameUser getOpponent() {
    // TODO: Preloading parent does not work. There is a db access on every parent loading 
    //       for quizzes added like this
    //       So I implemented this workaround to store parent locally
    if (this.opponent == null) {
      this.opponent = parent(OnlineGameUser.class);
    }
    return this.opponent;
  }

  public void setOpponent(final OnlineGameUser opponent) {
    this.opponent = opponent;
    setParent(opponent);
  }

  public OnlineGameResults getResults() {
    return this.results;
  }

  public void setResults(OnlineGameResults results) {
    this.results = results;
  }

  public QuizStatus getStatus() {
    final String statusString = getString(DB_STATUS);
    if (statusString == null) {
      return null;
    }
    return QuizStatus.valueOf(statusString);
  }

  public void setStatus(final QuizStatus status) {
    final QuizStatus oldStatus = getStatus();
    final QuizStatus newStatus = status;
    if ((oldStatus == null && newStatus == null) || (oldStatus != null && oldStatus.equals(newStatus)) ||
        (newStatus != null && newStatus.equals(oldStatus))) {
      return;
    }
    set(DB_STATUS, status.toString());
    logger.info("Quiz status geändert von " + oldStatus + " zu " + newStatus);
    notifyStatusListeners(oldStatus, newStatus);
  }
  
  public void setStatus(final NotificationType notificationType) {
    setStatus(getStatusFromNotificationType(notificationType));
  }
  
  public String getLastNotificationId() {
    return getString(DB_LAST_NOTIFICATION_ID);
  }
  
  public void setLastNotificationId(final String id) {
    set(DB_LAST_NOTIFICATION_ID, id);
  }
  
  public Timestamp getCreatedTimestamp() {
    return getTimestamp(DB_CREATED_AT);
  }
  
  @Override
  protected void afterSave() {
    super.afterSave();
    // refresh to retrieve default values of columns
    this.refresh();
  }
  
  public static QuizStatus getStatusFromNotificationType(final NotificationType notificationType) {
    
    switch (notificationType) {
    case CHALLENGE_INVITATION:
    case CHALLENGE_REJECTED:
    case CHALLENGE_ACCEPTED:
    case CHALLENGE_INVITATION_SENT:
    default:
      logger.error("Unaccaptable notification type!");
      
      return QuizStatus.FINISHED;
      
    case QUIZ_INVITATION:
      
      return QuizStatus.INVITATION;
      
    case QUIZ_INVITATION_SENT:

      return QuizStatus.INVITATION_SENT;

    case QUIZ_ACCEPTED:
      
      return QuizStatus.ACCEPTED;

    case QUIZ_REJECTED:
      
      return QuizStatus.REJECTED;

    case QUIZ_CHALLENGE_PENDING:
      
      return QuizStatus.PENDING;

    case SELECT_TOPIC:
      
      return QuizStatus.SELECT_TOPIC;

    case QUIZ_CHALLENGE_STARTED:
      
      return QuizStatus.STARTED;

    case QUIZ_CHALLENGE_ROUND_PLAYING:
      
      return QuizStatus.ROUND_PLAYING;

    case QUIZ_CHALLENGE_ROUND_WAITING:
      
      return QuizStatus.ROUND_WAITING;

    case QUIZ_CHALLENGE_WON:
      
      return QuizStatus.WON;

    case QUIZ_CHALLENGE_LOSE:
      
      return QuizStatus.LOST;

    case QUIZ_CHALLENGE_DRAW:
      
      return QuizStatus.DRAW;

    case FINISHED:
      
      return QuizStatus.FINISHED;

    }
    
  }
  
  private void notifyStatusListeners(final QuizStatus oldStatus, final QuizStatus newStatus) {
    checkForAddedOrRemoved();
    for (StatusListener listener: this.statusListeners) {
      listener.statusChanged(this, oldStatus, newStatus);
    }
    checkForAddedOrRemoved();
  }

  private void checkForAddedOrRemoved() {
    this.statusListeners.addAll(this.addedStatusListeners);
    this.addedStatusListeners.clear();
    this.statusListeners.removeAll(this.removedStatusListeners);
    this.removedStatusListeners.clear();
  }
  
  public void addStatusListener(final StatusListener listener) {
    this.addedStatusListeners.add(listener);
  }

  public void removeStatusListener(final StatusListener listener) {
    // This is a workaround to avoid ConcurrentModificationException on statusListeners
    // (removing while iterating in notifyStatusListeners)
    this.removedStatusListeners.add(listener);
  }

}
