package de.kreawi.trainer.onlinegame.net;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.commons.io.IOUtils;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.entity.mime.HttpMultipartMode;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.impl.client.HttpClients;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import de.kreawi.trainer.onlinegame.OnlineGameMain;
import de.kreawi.trainer.onlinegame.OnlineGameModel;
import de.kreawi.trainer.onlinegame.entities.OnlineGameProfile;
import de.kreawi.trainer.onlinegame.entities.OnlineGameServerError;
import de.kreawi.trainer.onlinegame.entities.OnlineGameUser;

public abstract class OnlineGameCommunicator {
  
  protected static final String ONLINE_GAME_PROTOCOL = "https://";
  protected static final String ONLINE_GAME_SERVER = "ikreawi.de";
  private static final String ONLINE_GAME_VERSION_PREFIX = "/v1"; 
  
  private static final int CONNECTION_TIMEOUT_MS = 20 * 1000; // Timeout in millis.

  public static final String RECORD_TYPE_QUIZ = "quiz";
  private static final String RECORD_TYPE_CHALLENGE = "challenge";  // not relevant for Trainer

  protected final OnlineGameModel model;

  private final Logger logger = LoggerFactory.getLogger(OnlineGameCommunicator.class);
  
  public OnlineGameCommunicator(final OnlineGameModel model) {
    this.model = model;
  }

  protected <T extends Object> T receiveJsonData(final String path, final HashMap<String, String> parameters,
      final Class<T> type) throws MalformedURLException, IOException, JsonParseException, JsonMappingException,
      OnlineGameCommunicatorException {
    
    if (!this.model.hasServerSession()) {
      requestSession();
    }
    
    final String url = createUrl(path, parameters);
    
    HttpClient httpclient = HttpClients.createDefault();
    HttpGet httpget = new HttpGet(url);
    httpget.setConfig(getRequestConfig());
    httpget.setHeader("session", this.model.getServerSession());
    
    HttpResponse response = httpclient.execute(httpget);
    
    final int statusCode = response.getStatusLine().getStatusCode();
    if (statusCode == 412) {
      // sesion expired -> renew
      
      requestSession();
      return receiveJsonData(path, parameters, type);
      
    } else if (statusCode != 200) {
      
      final OnlineGameServerError serverError = getServerError(response);
      final String errorDescription = createErrorDescription(serverError, response);
      
      throw new IllegalArgumentException("Wrong status code in response: " + statusCode + " (" + errorDescription + ")");
      
    }
    
    HttpEntity entity = response.getEntity();

    if (entity != null) {
        
        try (
            InputStream instream = entity.getContent();
            InputStreamReader isr = new InputStreamReader(instream);
            BufferedReader rd = new BufferedReader(isr);
            ) {
          
          final ObjectMapper jackson = new ObjectMapper();
          jackson.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
          
          return jackson.readValue(rd, type);
          
        }
    } else {
      throw new IllegalArgumentException("No response entity data!");
    }
    
  }

  private void requestSession() throws ClientProtocolException, IOException, OnlineGameCommunicatorException {
    
    final String oldSession = this.model.getServerSession();
    final String email = this.model.getSettings().getProfileUser().getEmail();
    final String jsonData = (new ObjectMapper()).writeValueAsString(
        new RequestSessionData(OnlineGameMain.APP_NAME, email, oldSession));
    final boolean appendSessionHeader = false;
    final String session = submitJsonData("/requestSession", jsonData, String.class, appendSessionHeader);
    
    this.model.setServerSession(session);
  }

  private static String createUrl(final String path, final HashMap<String, String> parameters) {
    String url = ONLINE_GAME_PROTOCOL + ONLINE_GAME_SERVER + ONLINE_GAME_VERSION_PREFIX + path;
    
    if (parameters != null && parameters.size() > 0) {
      String params = "";
      Iterator<String> it = parameters.keySet().iterator();
      while (it.hasNext()) {
        final String key = it.next();
        params += (params.length() > 0?"&":"?") + key + "=" + parameters.get(key);
      }
      url += params;
    }
    return url;
  }
  
  protected OnlineGameUser submitProfileData(final String urlPath, final OnlineGameProfile profile)
      throws ClientProtocolException, IOException, OnlineGameCommunicatorException {
    HttpClient httpclient = HttpClients.createDefault();
    HttpPost httppost = new HttpPost(createUrl(urlPath, null));  // parameters are in body not in address

    httppost.setConfig(getRequestConfig());
    httppost.setHeader("session", this.model.getServerSession());

    MultipartEntityBuilder builder = MultipartEntityBuilder.create();
    
    builder.setMode(HttpMultipartMode.BROWSER_COMPATIBLE);
    
    for (HashMap.Entry<String, String> entry : profile.getParameters().entrySet()) {
      if (entry.getValue() != null) {
        builder.addTextBody(entry.getKey(), entry.getValue(), ContentType.TEXT_PLAIN);
      }
    }
    
    final Path imagePath = profile.getProfileImage();
    
    if (imagePath != null && Files.exists(imagePath)) {
      builder.addBinaryBody(
          "profileImage",
          Files.newInputStream(imagePath),
          ContentType.APPLICATION_OCTET_STREAM,
          imagePath.getFileName().toString()
          );
    }

    HttpEntity multipart = builder.build();
    httppost.setEntity(multipart);
    
    logger.debug("Anzahl header: " + httppost.getAllHeaders().length);
    for (Header header : httppost.getAllHeaders()) {
      logger.debug(header.getName() + ": " + header.getValue());
    }
    
//    Header acceptHeader = new BasicHeader(HttpHeaders.ACCEPT, "application/json");
//    httppost.addHeader(acceptHeader);
//    Header contentTypeHeader = new BasicHeader(HttpHeaders.CONTENT_TYPE, "multipart/form-data");
//    httppost.addHeader(contentTypeHeader);

    //Execute and get the response.
    HttpResponse response = httpclient.execute(httppost);
    
    final int statusCode = response.getStatusLine().getStatusCode();
    if (statusCode != 200) {
      
      final OnlineGameServerError serverError = getServerError(response);
      final String errorDescription = createErrorDescription(serverError, response);
      
      logger.error("Error on submitting profile data!", 
          new IllegalArgumentException("Wrong status code in response: " + statusCode + " (" + errorDescription + ")"));

      throw new OnlineGameCommunicatorException(serverError);
      
    }
    
    HttpEntity entity = response.getEntity();

    if (entity != null) {
        
        try (
            InputStream instream = entity.getContent();
            InputStreamReader isr = new InputStreamReader(instream);
            BufferedReader rd = new BufferedReader(isr);
            ) {
          
          final ObjectMapper jackson = new ObjectMapper();
          jackson.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

          return jackson.readValue(rd, OnlineGameUser.class);

        }
    } else {
      throw new IllegalArgumentException("No response entity data!");
    }
    
  }
  
  protected <T extends Object> T submitJsonData(final String urlPath, final String jsonData,
      final Class<T> returnType) throws ClientProtocolException, IOException, OnlineGameCommunicatorException {
    return submitJsonData(urlPath, jsonData, returnType, true);
  }
  
  protected <T extends Object> T submitJsonData(final String urlPath, final String jsonData, final Class<T> returnType,
      final boolean appendSessionHeader) throws ClientProtocolException, IOException, OnlineGameCommunicatorException {
    
    if (appendSessionHeader && !this.model.hasServerSession()) {
      requestSession();
    }
    
    HttpClient httpclient = HttpClients.createDefault();
    HttpPost httppost = new HttpPost(createUrl(urlPath, null));  // parameters are in body not in address
    
    HttpEntity requestEntity = new StringEntity(jsonData, ContentType.APPLICATION_JSON);
    httppost.setEntity(requestEntity);
    httppost.setHeader("Content-Type", "application/json;charset=utf-8");

    httppost.setConfig(getRequestConfig());
    httppost.setHeader("session", this.model.getServerSession());

    //Execute and get the response.
    HttpResponse response = httpclient.execute(httppost);
    
    final int statusCode = response.getStatusLine().getStatusCode();

    if (statusCode == 412) {
      // sesion expired -> renew
      
      if (!appendSessionHeader) {
        final OnlineGameServerError serverError = getServerError(response);
        final String errorDescription = createErrorDescription(serverError, response);
        throw new IllegalArgumentException(
            "Status code " + statusCode + ", but session header should not be appended! Error: " + errorDescription);
      }
      requestSession();
      return submitJsonData(urlPath, jsonData, returnType);
      
    } else if (statusCode != 200) {
      
      final OnlineGameServerError serverError = getServerError(response);
      final String errorDescription = createErrorDescription(serverError, response);
      
      logger.error("Error on submitting json data!",
          new IllegalArgumentException("Wrong status code in response: " + statusCode + " (" + errorDescription + ")"));
      
      throw new OnlineGameCommunicatorException(serverError);
      
      
    }
    
    final HttpEntity entity = response.getEntity();
    
    if (entity != null && returnType != null) {
        
        try (
            InputStream instream = entity.getContent();
            InputStreamReader isr = new InputStreamReader(instream);
            BufferedReader rd = new BufferedReader(isr);
            ) {
          
          if (returnType == String.class) {
            return (T) IOUtils.toString(rd);
          } else {
            final ObjectMapper jackson = new ObjectMapper();
            jackson.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
            
            return jackson.readValue(rd, returnType);
          }

        }
    } else if (returnType != null) {
      throw new IllegalArgumentException("No response entity data! Should be: " + returnType.getName());
    } else {
      return null;    // no return type
    }
    
  }

  private String createErrorDescription(final OnlineGameServerError serverError, final HttpResponse response)
      throws IOException {
    
    final String errorDescription;
    if (serverError != null) {
      errorDescription = serverError.toString();
    } else {
      errorDescription = response.getStatusLine().getReasonPhrase();
    }
    return errorDescription;
  }

  private OnlineGameServerError getServerError(HttpResponse response) {
    HttpEntity entity = response.getEntity();
    OnlineGameServerError serverError = null;
    try (InputStream instream = entity.getContent();
         InputStreamReader isr = new InputStreamReader(instream);
         BufferedReader rd = new BufferedReader(isr);
        ) {
      
      final ObjectMapper jackson = new ObjectMapper();
      jackson.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

      serverError = jackson.readValue(rd, OnlineGameServerError.class);

    } catch (Exception e) {
      logger.error("Could not parse error content!", e);
    }
    return serverError;
  }
  
  private static RequestConfig getRequestConfig() {
    final RequestConfig requestConfig = RequestConfig.custom()
        .setConnectionRequestTimeout(CONNECTION_TIMEOUT_MS)
        .setConnectTimeout(CONNECTION_TIMEOUT_MS)
        .setSocketTimeout(CONNECTION_TIMEOUT_MS)
        .build();
    return requestConfig;
  }
  
  private class RequestSessionData {
    
    private final String appName;
    private final String email;
    private final String session;
    
    public RequestSessionData(String appName, String email, String session) {
      this.appName = appName;
      this.email = email;
      this.session = session;
    }

    public String getAppName() {
      return appName;
    }

    public String getEmail() {
      return email;
    }

    public String getSession() {
      return session;
    }
    
  }
  
  public static interface ServerFailureCallback {
    public void run(final String serverMessage);
  }
}
