package de.kreawi.trainer.onlinegame.net;

import java.nio.file.Path;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.kreawi.trainer.onlinegame.OnlineGameModel;
import de.kreawi.trainer.onlinegame.entities.OnlineGameLoginResponse;
import de.kreawi.trainer.onlinegame.entities.OnlineGameProfile;
import de.kreawi.trainer.onlinegame.entities.OnlineGameUser;

public class OnlineGameProfileCommunicator extends OnlineGameCommunicator {

  private final Logger logger = LoggerFactory.getLogger(OnlineGameProfileCommunicator.class);
  
  public OnlineGameProfileCommunicator(final OnlineGameModel model) {
    super(model);
  }

  public void signup(final OnlineGameUser profileUser, final String password, final Path profileImage,
      final ProfileUserDataReceiver onSuccess, final ServerFailureCallback onFailure) {
    
    final OnlineGameProfile profile = new OnlineGameProfile(profileUser, password, profileImage);    

    logger.info("Signing up profile: " + profile.toString());
    
    try {
      final OnlineGameUser userData = submitProfileData("/signup", profile);
      onSuccess.run(userData);
    } catch (OnlineGameCommunicatorException e) {
      onFailure.run(e.getServerMessage());
    } catch (Exception e) {
      onFailure.run("");
    }
    
  }
  
  public void editProfile(final OnlineGameUser profileUser, final String password, final Path profileImage,
      final ProfileUserDataReceiver onSuccess, final ServerFailureCallback onFailure) {
    
    final OnlineGameProfile profile = new OnlineGameProfile(profileUser, password, profileImage);    
    
    try {
      
      final OnlineGameUser userData = submitProfileData("/editUserProfile", profile);

      onSuccess.run(userData);

    } catch (OnlineGameCommunicatorException e) {
      onFailure.run(e.getServerMessage());
    } catch (Exception e) {
      onFailure.run("");
    }
    
  }
  
  public void login(final String email, final String password, final ProfileUserDataReceiver onSuccess,
      final ServerFailureCallback onFailure) {
    
    final LoginData loginData = new LoginData(email, password, null, null);

    try {
      
      final boolean appendSessionHeader = false;
      final OnlineGameLoginResponse loginResponse = submitJsonData("/login",
          (new ObjectMapper()).writeValueAsString(loginData), OnlineGameLoginResponse.class, appendSessionHeader);
      
      this.model.setServerSession(loginResponse.getSession());
      
      logger.debug("-------- loginResponse.toString(): " + loginResponse.toString());
      
      onSuccess.run(loginResponse.getUserShell());
      
    } catch (OnlineGameCommunicatorException e) {
      onFailure.run(e.getServerMessage());
    } catch (Exception e) {
      logger.error("Error on login!", e);
      onFailure.run("");
    }
    
  }
  
  public void logout(final OnlineGameUser profileUser, final Runnable onSuccess, final Runnable onFailure) {

    final LogoutData logoutData = new LogoutData(profileUser.getId(), profileUser.getDeviceToken());

    try {
      
      submitJsonData("/logout", (new ObjectMapper()).writeValueAsString(logoutData), null);
      
      onSuccess.run();
      
    } catch (Exception e) {
      logger.error("Error on logout!", e);
      onFailure.run();
    }
    
  }
  
  public void removeFriend(final OnlineGameUser profileUser, final OnlineGameUser friend, Runnable onSuccess, Runnable onFailure) {
    
    final RemoveFriendData removeFriendData = new RemoveFriendData(profileUser.getId(), friend.getId());

    try {
      
      submitJsonData("/removeFriend", (new ObjectMapper()).writeValueAsString(removeFriendData), null);
      
      onSuccess.run();
      
    } catch (Exception e) {
      logger.error("Error on removing friend!", e);
      onFailure.run();
    }

  }

  public void changeStatus(final ChangeStatusData changeStatusData, final Runnable onSuccess,
      final Runnable onFailure) {
    
    try {
      
      submitJsonData("/changeOpenList", (new ObjectMapper()).writeValueAsString(changeStatusData), null);
      
      onSuccess.run();
      
    } catch (Exception e) {
      logger.error("Error on changing online status!", e);
      onFailure.run();
    }
    
  }
  
  public class LoginData {
   
    public LoginData(String email, String password, String appRecordId, String appName) {
      this.email = email;
      this.password = password;
      this.appRecordId = appRecordId;
      this.appName = appName;
    }

    private String email;
    
    private String password;
    
    private String appRecordId;
    
    private String appName;

    public String getEmail() {
      return email;
    }

    public void setEmail(String email) {
      this.email = email;
    }

    public String getPassword() {
      return password;
    }

    public void setPassword(String password) {
      this.password = password;
    }

    public String getAppRecordId() {
      return appRecordId;
    }

    public void setAppRecordId(String appRecordId) {
      this.appRecordId = appRecordId;
    }

    public String getAppName() {
      return appName;
    }

    public void setAppName(String appName) {
      this.appName = appName;
    }
    
  }
  
  public static class LogoutData {
    
    private String userId;
    
    private String deviceToken;

    public LogoutData(String userId, String deviceToken) {
      this.userId = userId;
      this.deviceToken = deviceToken;
    }

    public String getUserId() {
      return userId;
    }

    public void setUserId(String userId) {
      this.userId = userId;
    }

    public String getDeviceToken() {
      return deviceToken;
    }

    public void setDeviceToken(String deviceToken) {
      this.deviceToken = deviceToken;
    }
    
  }
  
  public static class RemoveFriendData {

    private String userId;
    private String friendId;

    public RemoveFriendData(final String userId, String friendId) {
      this.setUserId(userId);
      this.setFriendId(friendId);
    }

    public String getUserId() {
      return userId;
    }

    public void setUserId(String userId) {
      this.userId = userId;
    }

    public String getFriendId() {
      return friendId;
    }

    public void setFriendId(String friendId) {
      this.friendId = friendId;
    }
    
  }
  
  public static class ChangeStatusData {
    
    private String userId;
    
    private boolean open;

    public ChangeStatusData(String userId, boolean open) {
      this.userId = userId;
      this.open = open;
    }

    public String getUserId() {
      return userId;
    }

    public void setUserId(String userId) {
      this.userId = userId;
    }

    public boolean isOpen() {
      return open;
    }

    public void setOpen(boolean open) {
      this.open = open;
    }
    
  }
  
  public interface ProfileUserDataReceiver {
    public void run(OnlineGameUser userData);
  }

}
