package de.kreawi.trainer.update.autoupdate;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.NoSuchAlgorithmException;
import java.util.Observable;
import java.util.Observer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.SwingWorker;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.KreawiGlobals;
import de.kreawi.trainer.license.LicenseFileHandler;
import de.kreawi.trainer.update.Updater;
import de.kreawi.trainer.util.OSDetector;
import de.kreawi.trainer.util.TrainerTextResources;
import de.kreawi.trainer.util.swingx.DownloaderWithProgress;
import de.martinknauer.util.Util;
import de.martinknauer.util.Version;
import de.martinknauer.util.progress.ProgressInputStream;
import de.martinknauer.util.progress.ProgressListener;

public class AutoUpdater {

	public static final String VERSION_TAG = "version";

	public static final String UPDATER_URL_TAG = "updaterUrl";

  public static final String MAC_UPDATER_URL_TAG = "macUpdaterUrl";

	public static final String URL_TAG = "url";

	public static final String SIZE_TAG = "size";
	
	public static final String MESSAGE_TAG = "message";
	
	public static final String BULLETIN_TIMESTAMP_TAG = "bulletinTimestamp";
	
	public static final String BULLETIN_MESSAGE_TAG = "bulletinMessage";

	public static final String AUTO_UPDATER_JAR_NAME = "updater.jar";
	
	public static final String MAC_UPDATER_ZIP_NAME = "TrainerMacAutoUpdater.tar.gz";
	
	public static final String MAC_UPDATER_APP_NAME = "KreawiTrainerUpdater.app";

  public static final String AUTO_UPDATER_ADMIN_EXE_NAME = "UpdaterAdmin.exe";

	private static final String SID_PLACEHOLDER = "\\{sid\\}";
	
	private static final String OS_PLACEHOLDER = "\\{os\\}";
	
	private static final String VERSION_PLACEHOLDER = "\\{version\\}";

	private URL autoUpdateUrl;

	private Version updateVersion;

	private URL updateUrl;

	private int updateSize;
	
	private String updateMessage;

	private boolean isAutoUpdateAvailable = true;

	private URL updateUpdaterUrl;

	private String bulletinTimestamp;
	
	private String bulletinMessage;
	
	private final boolean updateFullVersionData;

	private static final Logger logger = LoggerFactory.getLogger(AutoUpdater.class); 

	public AutoUpdater(final boolean updateFullVersionData) {
	  this.updateFullVersionData = updateFullVersionData;
		this.autoUpdateUrl = getAutoUpdateUrl(updateFullVersionData);
		logger.info("AutoUpdater initialized with url " + autoUpdateUrl);
	}
	
	private static URL getAutoUpdateUrl(final boolean updateFullVersionData) {
	  String updateUrl = "";
    try {
      if (updateFullVersionData) {
        updateUrl = TrainerTextResources.getStringWithSubstitution("AUTO_UPDATE_URL_FULL_VERSION_DATA",
            KreawiGlobals.CURRENT_YEAR);
      } else {
        updateUrl = TrainerTextResources.getStringWithSubstitution("AUTO_UPDATE_URL", KreawiGlobals.CURRENT_YEAR);
      }
      return new URL(updateUrl);
    } catch (MalformedURLException e) {
      logger.error("WARNING: Problem while parsing URL '" + updateUrl + "'.", e);
      return null;
    }
  }
  
	public void retrieveUpdateInfos() throws IOException {

		InputStream urlStream;
    try {
      urlStream = this.autoUpdateUrl.openStream();
    } catch (IOException e) {
      logger.info("Konnte Update-Infos nicht laden! (" + e.toString() + ")");
      this.isAutoUpdateAvailable = false;
      return;
    }

		// read in entire file
		byte b[] = new byte[1000];
		int numRead = urlStream.read(b);
		String htmlString = new String(b, 0, numRead);
		while (numRead != -1) {
			numRead = urlStream.read(b);
			if (numRead != -1) {
				String newHtmlString = new String(b, 0, numRead);
				htmlString += newHtmlString;
			}
		}

		try {
      this.updateVersion = new Version(getItemFromHTML(htmlString, VERSION_TAG));
		} catch (IllegalArgumentException e) {
			throw new IOException();
		}
		
		logger.info("Auto-update version is " + this.updateVersion);
		
		try {
	    if (OSDetector.getOs().isMac()) {
	      this.updateUpdaterUrl = new URL(getItemFromHTML(htmlString, MAC_UPDATER_URL_TAG));
	    } else {
        this.updateUpdaterUrl = new URL(getItemFromHTML(htmlString, UPDATER_URL_TAG));
	    }
		} catch (MalformedURLException e) {
		  logger.info("Konnte Updater-Url nicht lesen! (" + e.toString() + ")");
			this.isAutoUpdateAvailable = false;
		}

		try {
			
			String urlString = getItemFromHTML(htmlString, URL_TAG);
			
			if (urlString != null) {

				String sid = LicenseFileHandler.getCurrentLicenceKey();
				if (sid == null) {
					sid = "";
				} else {
					try {
						sid = Util.md5Hash(sid);
					} catch (NoSuchAlgorithmException e) {
						sid = null;
					} catch (IOException e) {
						sid = null;
					}
				}

				String os = System.getProperty("os.name") + " "
						+ System.getProperty("os.arch") + " "
						+ System.getProperty("os.version");
				os = URLEncoder.encode(os, "UTF-8");

				Version version = Util.getVersion();
				String versionStr = "";
				if (version != null) {
					versionStr = URLEncoder.encode(version.toString(), "UTF-8");
				}

				urlString = urlString.replaceAll(SID_PLACEHOLDER, sid);
				urlString = urlString.replaceAll(OS_PLACEHOLDER, os);
				urlString = urlString.replaceAll(VERSION_PLACEHOLDER,
						versionStr);

				this.updateUrl = new URL(urlString);

				logger.info("Auto-update content URL is " + this.updateUrl);

			} else {
			  logger.info("Konnte Updater-Url nicht lesen! (urlString ist NULL)");
	      this.isAutoUpdateAvailable = false;
			}
			
		} catch (MalformedURLException e) {
		  logger.info("Konnte Updater-Url nicht lesen! (" + e.toString() + ")");
      this.isAutoUpdateAvailable = false;
		}

		String sizeString = getItemFromHTML(htmlString, SIZE_TAG);

		if (sizeString != null) {
		  try {
		    this.updateSize = Integer.parseInt(sizeString.trim());
		  } catch (NumberFormatException e) {
		    logger.info("Update-Größe kann nicht als zahl geparst werden! (sizeString: '" + sizeString + "')");
		  }
		}
		
		this.updateMessage = getItemFromHTML(htmlString, MESSAGE_TAG);
		
    this.bulletinTimestamp = getItemFromHTML(htmlString, BULLETIN_TIMESTAMP_TAG);
    
		this.bulletinMessage = getItemFromHTML(htmlString, BULLETIN_MESSAGE_TAG);

	}
  
	public File performAutoupdaterUpdate() throws IOException {
	  return performAutoupdaterUpdate(null);
	}

	public File performAutoupdaterUpdate(final SwingWorker swingWorker) throws IOException {

		if (this.isAutoUpdateAvailable) {
		  logger.info("Downloading updater from " + this.updateUpdaterUrl);

      final Path outdir = Files.createTempDirectory("kreawiUpdate");
			
      final File outfile;

      if (OSDetector.getOs().isMac()) {
			  outfile = new File(outdir.toFile(), MAC_UPDATER_ZIP_NAME);
			} else {
			  outfile = new File(outdir.toFile(), AUTO_UPDATER_JAR_NAME);
			}
      
      logger.info("   ... to " + outfile.getAbsolutePath());
			
      final DownloaderWithProgress downloader = new DownloaderWithProgress(updateUpdaterUrl, outfile);
      if (swingWorker != null) {
        downloader.addObserver(new Observer() {
          public void update(Observable o, Object arg) {
            final int oldProgress = swingWorker.getProgress();
            final int newProgress = (int) (downloader.getProgress() * 0.97f);
            if (oldProgress != newProgress) {
              if (KreawiGlobals.isDevMode()) {
                logger.info("Updating progress to " + newProgress);
              }
              swingWorker.firePropertyChange("progress", swingWorker.getProgress(), newProgress);
            }
          }
        });
      }
      
      downloader.download();
      
      while (downloader.getStatus() == DownloaderWithProgress.DOWNLOADING) {
        try {
          Thread.sleep(10);
        } catch (InterruptedException e) {
          logger.error("Download-Fehler! ", e);
        }
      }
			
      logger.info("Updater download finished.");

			return outfile;
		}

		return null;
	}

	public void downloadUpdate(ProgressListener progressListener) throws IOException {

		if (this.isAutoUpdateAvailable) {

			final ProgressInputStream urlStream = new ProgressInputStream(this.updateUrl.openStream());
			urlStream.addProgressListener(progressListener);
			urlStream.setSize(updateSize);
			
			final File outfile = new File(System.getProperty("user.dir"), Updater.FILE_NAME_UPDATE_ZIP);
			logger.info("Downloading update date file to '" + outfile.getPath() + "'...");
      Util.copyStream(urlStream, new FileOutputStream(outfile));
			
      logger.info("...finshed!");

		}

	}

	private String getItemFromHTML(String htmlString, String itemTag) {
		try {
			String regex = "(autoupdate\\." + itemTag + ":)(.+?)(:" + itemTag
					+ "\\.autoupdate)";
			Matcher matcher = Pattern.compile(regex).matcher(htmlString);

			if (matcher.find()) {
				return matcher.group(2);
			}
		} catch (IllegalStateException ise) {
			/* do nothing */
		}

		return "";

	}

	public static boolean hasAutoUpdatePermissions() {
    return Util.canWriteInDirectory(new File(System.getProperty("user.dir")));
	}

	public String getUpdateMessage() {
		return updateMessage;
	}
	
	public String getBulletinTimestamp() {
		return bulletinTimestamp;
	}
	
	public String getBulletinMessage() {
		return bulletinMessage;
	}

	public int getUpdateSize() {
		return updateSize;
	}

	public URL getUpdateUrl() {
		return updateUrl;
	}

	public Version getUpdateVersion() {
		return updateVersion;
	}

	public boolean isAutoUpdateAvailable() {
		return isAutoUpdateAvailable;
	}

  public boolean isUpdateFullVersionData() {
    return updateFullVersionData;
  }
	
}
