package de.kreawi.trainer.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import javax.swing.JEditorPane;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLFrameHyperlinkEvent;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.KreawiGlobals;
import de.kreawi.trainer.TrainerFrame;
import de.kreawi.trainer.util.swingx.CursorToolkit;
import de.kreawi.trainer.util.swingx.DesktopApi;
import de.kreawi.trainer.viewer.AudioViewer;
import de.kreawi.trainer.viewer.ImageViewer;
import de.kreawi.trainer.viewer.VideoViewer;
import de.kreawi.trainer.viewer.VideoViewerException;
import de.kreawi.trainer.viewer.ViewerException;
import edu.stanford.ejalbert.BrowserLauncher;

public class TrainerHyperlinkListener implements HyperlinkListener {

  private TrainerFrame mainframe;

	private boolean imageViewerEnabled;
	private ImageViewer imageViewer;
	private boolean audioViewerEnabled;
	private AudioViewer audioViewer;
  private VideoViewer videoViewer;

  protected final long TIMEOUT_VIDEO_PLAYBACK_MILLIS = 10000;

  private static final Logger logger = LoggerFactory.getLogger(TrainerHyperlinkListener.class); 

	public TrainerHyperlinkListener(TrainerFrame mainframe) {
	  this.mainframe = mainframe;
		this.imageViewerEnabled = true;
		this.audioViewerEnabled = true;
	}

	public void hyperlinkUpdate(HyperlinkEvent e) {

		if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {

		  logger.info("TrainerHyperlinkListener.hyperlinkUpdate() called for URL: " + e.getURL());
		  
			final URL url;
			if (e.getURL().getProtocol().equals("file")) {
			  
			  try {
			    final URI uri = e.getURL().toURI();
          final String urlPath = uri.getSchemeSpecificPart().replace('/', File.separatorChar);
			    if (new File(urlPath).isAbsolute()) {
			      url = uri.toURL();
			    } else {
			      final File workDir = new File(System.getProperty("user.dir"));
			      final File resourcesDir;
			      if (KreawiGlobals.isDemo()) {
			        resourcesDir = new File(workDir, KreawiGlobals.DEMO_DIR_NAME + File.separator);
			      } else {
			        resourcesDir = new File(workDir, KreawiGlobals.RESOURCES_DIR_NAME + File.separator);
			      }
            final File file = new File(resourcesDir, urlPath);
			      if (file.exists()) {
			        url = file.toURI().toURL();
			        logger.info("url changed to '" + url.toString() + "'!");
			      } else {
			        logger.info("File not found: " + e.getURL().toString());
			        displayErrorImage();
			        return;
			      }
			    }
			  } catch (MalformedURLException e1) {
			    logger.error("Error reading URL!", e1);
			    displayErrorImage();
			    return;
			  } catch (URISyntaxException e2) {
			    logger.error("Error reading URL!", e2);
          displayErrorImage();
          return;
			  }
			} else {
			  url = e.getURL();
			}
			

			try {
			  SwingUtilities.invokeLater(new Runnable() {
          public void run() {
            CursorToolkit.startWaitCursor(mainframe.getRootPane());
          }
        });

				if (e instanceof HTMLFrameHyperlinkEvent) {
					HTMLFrameHyperlinkEvent evt = (HTMLFrameHyperlinkEvent) e;
					HTMLDocument doc = (HTMLDocument) e.getSourceElement().getDocument();
					doc.processHTMLFrameHyperlinkEvent(evt);
				} else {

					String s = " ";
					try {
						s = e.getSourceElement().getDocument()
								.getText(
										e.getSourceElement().getStartOffset(),
										e.getSourceElement().getEndOffset()
												- e.getSourceElement()
														.getStartOffset());
					} catch (BadLocationException ble) {
						// do nothing
					} catch (NullPointerException npe) {
					  // do nothing
					}
					
					final String linkedText = s;

					if (ImageViewer.isImageUrl(url)
							&& this.imageViewerEnabled) {
						if (this.imageViewer != null) {
							imageViewer.setVisible(false);
							imageViewer.dispose();
						}

						try {
							this.imageViewer = new ImageViewer(mainframe,
									linkedText, url);
						} catch (ViewerException ve) {
							handleEventExternally(url);
						}
					} else if (VideoViewer.isVideoUrl(url)) {
					  if (this.videoViewer != null) {
					    videoViewer.setVisible(false);
              videoViewer.dispose();
            }
					  final SwingWorker<VideoViewer, Void> worker = new SwingWorker<VideoViewer, Void>() {
              @Override
              protected VideoViewer doInBackground() throws Exception {
                return new VideoViewer(mainframe, linkedText, url);
              }
              
              @Override
              protected void done() {
                try {
                  videoViewer = get(TIMEOUT_VIDEO_PLAYBACK_MILLIS, TimeUnit.MILLISECONDS);
                  videoViewer.setVisible(true);
                } catch (InterruptedException e) {
                  logger.error("Laden des Videos unterbrochen!", e);
                } catch (ExecutionException e) {
                  final String errorMessage;
                  final Throwable causeException = e.getCause();
                  if (causeException instanceof VideoViewerException) {
                    errorMessage = causeException.getMessage();
                  } else {
                    logger.error("Unerwarteter Fehler beim Anzeigen des Videos!", causeException);
                    errorMessage = TrainerTextResources.getString("VIDEO_VIEWER_ERROR_UNKNOWN");
                  }
                  mainframe.showMessageDialog(errorMessage, JOptionPane.WARNING_MESSAGE);
                } catch (CancellationException e) {
                  logger.info("Laden des Videos wurde vom Benutzer abgebrochen!");
                } catch (TimeoutException e) {
                  logger.info("Timeout beim Anzeigen des Videos!");
                  e.printStackTrace();
                }
              }
					  };
					  final String waitMessage = TrainerTextResources.getString("VIDEO_VIEWER_LOADING");
            this.mainframe.showPleaseWaitDialog(waitMessage, worker);
					  
					} else if (AudioViewer.isAudioUrl(url)
							&& this.audioViewerEnabled) {

						String title = mainframe.getTitle();
						if (title != null) {
							title += " ("
									+ TrainerTextResources
											.getString("QUIZ_TITLE_AUDIO_SUFFIX")
									+ ")";
						} else {
							title = " ";
						}

						if (this.audioViewer != null) {
							audioViewer.setVisible(false);
							audioViewer.dispose();
						}

						try {
							this.audioViewer = new AudioViewer(mainframe, title, e
									.getURL());
						} catch (ViewerException ve) {
							handleEventExternally(url);
						}

					} else if (url.toString().startsWith("http") && url.toString().endsWith(".pdf")) {
					  // handle external PDF file links from online game
					  final File pdfFile = downloadPdfToTemp(url);
					  if (pdfFile == null || !pdfFile.exists()) {
					    logger.error("Could not download PDF file!");
					    displayErrorImage();
					    return;
					  }
					  try {
              handleEventExternally(pdfFile.toURI().toURL());
            } catch (MalformedURLException e1) {
              logger.error("Error converting file location to URL!", e1);
              displayErrorImage();
            }
					} else {
						handleEventExternally(url);
					}

				}
			} finally {
				Timer timer = new Timer();
				timer.schedule(new TimerTask() {
					@Override
					public void run() {
					  SwingUtilities.invokeLater(new Runnable() {
					    public void run() {
					      CursorToolkit.stopWaitCursor(mainframe.getRootPane());
					    }
					  });
					}
				}, 1000);
			}
		}

	}
	
  private File downloadPdfToTemp(URL url) {
    
    try {
      final File localFile = File.createTempFile("trainer-", ".pdf");
      localFile.deleteOnExit();
      
      try (
          ReadableByteChannel rbc = Channels.newChannel(url.openStream());
          FileOutputStream fos = new FileOutputStream(localFile);
          ) {
        fos.getChannel().transferFrom(rbc, 0, Long.MAX_VALUE);
      }
      
      return localFile;
    } catch (IOException e) {
      logger.error("Failure while downloading PDF file!", e);
    }
    
    return null;
  }

  private void handleEventExternally(final URL url) {
    try {
      if (url.toString().startsWith("http")) {
        DesktopApi.browse(url.toURI());
      } else {
        DesktopApi.open(new File(url.toURI().getPath()));
      }
    } catch (Exception ex) {
      logger.error("Desktop.browse hat nicht funktioniert! (url: " + url.toString() + ")", ex);
      try {
        (new BrowserLauncher()).openURLinBrowser(url.toString());
      } catch (Exception ex2) {
        logger.error("Konnte URL nicht öffnen! (url: " + url.toString() + ")", ex2);
        displayErrorImage();
      }
    }
  }
  
	private void displayErrorImage() {
		try {
			URL url = new File(new File(System.getProperty("user.dir"),
					KreawiGlobals.DEMO_DIR_NAME),
					KreawiGlobals.ERROR_IMAGE_FILE_NAME).toURI().toURL();
      this.imageViewer = new ImageViewer(mainframe, TrainerTextResources.getString("ERROR"), url);
		} catch (MalformedURLException e) {
		  logger.error("Unable to display error image!", e);
		} catch (ViewerException e) {
		  logger.error("Unable to display error image!", e);
		}
	}

	public boolean isImageViewerEnabled() {
		return imageViewerEnabled;
	}

	public void setImageViewerEnabled(boolean imageViewerEnabled) {
		this.imageViewerEnabled = imageViewerEnabled;
	}

	public boolean isAudioViewerEnabled() {
		return audioViewerEnabled;
	}

	public void setAudioViewerEnabled(boolean audioViewerEnabled) {
		this.audioViewerEnabled = audioViewerEnabled;
	}

}
