package de.kreawi.trainer.viewer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.concurrent.TimeUnit;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JToolBar;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.KreawiGlobals;
import de.kreawi.trainer.dialogs.KreawiInternalFrame;
import de.kreawi.trainer.util.OSDetector;
import de.kreawi.trainer.util.TrainerTextResources;
import uk.co.caprica.vlcj.component.DirectMediaPlayerComponent;
import uk.co.caprica.vlcj.discovery.NativeDiscovery;
import uk.co.caprica.vlcj.player.MediaPlayer;
import uk.co.caprica.vlcj.player.MediaPlayerEventAdapter;
import uk.co.caprica.vlcj.player.direct.BufferFormat;
import uk.co.caprica.vlcj.player.direct.BufferFormatCallback;
import uk.co.caprica.vlcj.player.direct.DirectMediaPlayer;
import uk.co.caprica.vlcj.player.direct.RenderCallback;
import uk.co.caprica.vlcj.player.direct.RenderCallbackAdapter;
import uk.co.caprica.vlcj.player.direct.format.RV32BufferFormat;

public class VideoViewer extends ViewerWindow {
  
  private static final double SLIDER_STEP_SIZE = 1000.0;

  private static int videoWidth = 600;

  private static int videoHeight = 450;
  
  private final static int TOOLBAR_HEIGHT = 50;
  
  private final VideoSurfacePanel videoSurface;

  private final BufferedImage image;

  private final DirectMediaPlayerComponent mediaPlayerComponent;
  
  private boolean scaled = false;

  private JComponent mediaControlPanel;

  private final Dimension maximumVideoDimension;

  private JSlider slider;

  private JButton playButton;

  private JLabel timeLabel;
  
  private static final Logger logger = LoggerFactory.getLogger(VideoViewer.class); 

  public VideoViewer(Component owner, String title, final URL url) throws VideoViewerException, ViewerException {
    super(owner, title);
    
    final JPanel gradientPanel = new KreawiInternalFrame("", new BorderLayout()).getContent();
    
    final boolean vlcFound = new NativeDiscovery().discover();
    if (!vlcFound) {
      final String errorMessage;
      final String bitSize = System.getProperty("sun.arch.data.model");
      if (!OSDetector.getOs().isMac() && bitSize != null && "32".equals(bitSize) || "64".equals(bitSize)) {
        errorMessage = TrainerTextResources.getStringWithSubstitution(
            "VIDEO_VIEWER_ERROR_VLC_PLAYER_NOT_FOUND_WITH_BIT_SIZE", bitSize);
      } else {
        errorMessage = TrainerTextResources.getString("VIDEO_VIEWER_ERROR_VLC_PLAYER_NOT_FOUND");
      }
      throw new VideoViewerException(errorMessage);
    }

    maximumVideoDimension = new Dimension(getOwner().getWidth(),
        getOwner().getHeight() - TOOLBAR_HEIGHT - 30);

    videoSurface = new VideoSurfacePanel();
    gradientPanel.add(videoSurface, BorderLayout.CENTER);
    BufferFormatCallback bufferFormatCallback = new BufferFormatCallback() {
        public BufferFormat getBufferFormat(int sourceWidth, int sourceHeight) {
          return new RV32BufferFormat(maximumVideoDimension.width, maximumVideoDimension.height);
        }
    };
    mediaPlayerComponent = new DirectMediaPlayerComponent(bufferFormatCallback) {
        @Override
        protected RenderCallback onGetRenderCallback() {
            return new TutorialRenderCallbackAdapter();
        }
    };
    addWindowListener(new WindowAdapter() {
      public void windowClosed(WindowEvent e) {
        mediaPlayerComponent.release();
      }
    });
    mediaPlayerComponent.getMediaPlayer().addMediaPlayerEventListener(new MediaPlayerEventAdapter() {

      public void videoOutput(MediaPlayer mediaPlayer, int newCount) {
        if (!scaled) {
          final Dimension dimension = scaleVideo(mediaPlayer.getVideoDimension());
          videoWidth=dimension.width;
          videoHeight=dimension.height;
          videoSurface.refreshVideoSize();
          mediaControlPanel.setPreferredSize(new Dimension(videoWidth, TOOLBAR_HEIGHT));
          VideoViewer.this.pack();
          scaled=true;
        }
      }
      public void timeChanged(MediaPlayer mediaPlayer, long newTime) {
        
        refreshTimeLabel(newTime, mediaPlayer.getLength());
        
        final int newSliderValue = (int) (((double) newTime) / ((double) mediaPlayer.getLength()) * SLIDER_STEP_SIZE);
        slider.setValue(newSliderValue);
      }
      public void mediaStateChanged(MediaPlayer mediaPlayer, int newState) {
        final ImageIcon defaultPlayPauseIcon;
        final ImageIcon playPauseIconRollover;
        final ImageIcon playPauseIconPressed;
        if (mediaPlayer.isPlaying()) {
          String pauseIconUrl= KreawiGlobals.APP_IMAGES_DIR_NAME + "/pause28_grey.png";
          defaultPlayPauseIcon = new ImageIcon(KreawiGlobals.class
              .getResource(pauseIconUrl));
          
          String pauseIconUrlRollover = KreawiGlobals.APP_IMAGES_DIR_NAME + "/pause28_grey_light.png";
          playPauseIconRollover = new ImageIcon(KreawiGlobals.class
              .getResource(pauseIconUrlRollover));
          
          String pauseIconUrlPressed = KreawiGlobals.APP_IMAGES_DIR_NAME + "/pause28_grey_dark.png";
          playPauseIconPressed = new ImageIcon(KreawiGlobals.class
              .getResource(pauseIconUrlPressed));
        } else {
          String playIconUrl = KreawiGlobals.APP_IMAGES_DIR_NAME + "/play28_grey.png";
          defaultPlayPauseIcon = new ImageIcon(KreawiGlobals.class
              .getResource(playIconUrl));
          
          String playIconUrlRollover = KreawiGlobals.APP_IMAGES_DIR_NAME + "/play28_grey_light.png";
          playPauseIconRollover = new ImageIcon(KreawiGlobals.class
              .getResource(playIconUrlRollover));
          
          String playIconUrlPressed = KreawiGlobals.APP_IMAGES_DIR_NAME + "/play28_grey_dark.png";
          playPauseIconPressed = new ImageIcon(KreawiGlobals.class
              .getResource(playIconUrlPressed));
        }
        playButton.setIcon(defaultPlayPauseIcon);
        playButton.setRolloverIcon(playPauseIconRollover);
        playButton.setPressedIcon(playPauseIconPressed);
      }
      public void error(MediaPlayer mediaPlayer) {
        logger.error("Fehler im VLC Media Player!");
      }
    });
    mediaControlPanel = createMediaControlPanel(mediaPlayerComponent.getMediaPlayer(), maximumVideoDimension.width);
    gradientPanel.add(mediaControlPanel, BorderLayout.SOUTH);
    image = GraphicsEnvironment
      .getLocalGraphicsEnvironment()
      .getDefaultScreenDevice()
      .getDefaultConfiguration()
      .createCompatibleImage(maximumVideoDimension.width, maximumVideoDimension.height);
    sif.add(gradientPanel);
    setBounds(100, 100, videoWidth, videoHeight);
    if (!canConnectToVideo(url) || !mediaPlayerComponent.getMediaPlayer().prepareMedia(url.toString())) {
      final String message = TrainerTextResources.getString("VIDEO_VIEWER_ERROR_FILE_NOT_PREPARED");
      throw new VideoViewerException(message );
    }
  }
  
  public void setVisible(boolean visible) {
    super.setVisible(visible);
    if (visible) {
      mediaPlayerComponent.getMediaPlayer().play();
    }
  }

  private JComponent createMediaControlPanel(final MediaPlayer mediaPlayer, final int initialWidth) {
    String pauseIconUrl = KreawiGlobals.APP_IMAGES_DIR_NAME + "/pause28_grey.png";
    final ImageIcon pauseIcon = new ImageIcon(KreawiGlobals.class
        .getResource(pauseIconUrl));
    
    String pauseIconUrlRollover = KreawiGlobals.APP_IMAGES_DIR_NAME + "/pause28_grey_light.png";
    final ImageIcon pauseIconRollover = new ImageIcon(KreawiGlobals.class
        .getResource(pauseIconUrlRollover));
    
    String pauseIconUrlPressed = KreawiGlobals.APP_IMAGES_DIR_NAME + "/pause28_grey_dark.png";
    final ImageIcon pauseIconPressed = new ImageIcon(KreawiGlobals.class
        .getResource(pauseIconUrlPressed));

    playButton = createRolloverButton(pauseIcon,
        pauseIconRollover, pauseIconPressed, new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            if (mediaPlayer.isPlaying()) {
              mediaPlayer.pause();
            } else if (mediaPlayer.isMediaParsed()){
              mediaPlayer.play();
            } else {
              logger.info("Play / pause button pressed before video info was parsed.");
            }
          }
        });

    // Default icon for JSlider changed in TrainerMain.main()
    slider = new JSlider(0, 1000, 0);
    slider.setOpaque(false);
    slider.addChangeListener(new ChangeListener() {
      private boolean paused;

      public void stateChanged(ChangeEvent e) {
        if (slider.getValueIsAdjusting()) {
          if (mediaPlayer.isPlaying()) {
            mediaPlayer.pause();
            paused = true;
          }
        } else if (!mediaPlayer.isPlaying()) {
          final int newTime = (int) (((double) slider.getValue()) / SLIDER_STEP_SIZE * ((double) mediaPlayer.getLength()));
          mediaPlayer.setTime(newTime);
          refreshTimeLabel(newTime, mediaPlayer.getLength());
          if (paused) {
            mediaPlayer.play();
            paused = false;
          }
        }
      }
    });
    
    timeLabel = new JLabel("0:00/0:00");
    
    JPanel leftGluePanel = new JPanel();
    leftGluePanel.setBorder(BorderFactory.createEmptyBorder());
    leftGluePanel.setPreferredSize(new Dimension(5, 5));
    leftGluePanel.setMaximumSize(leftGluePanel.getPreferredSize());
    leftGluePanel.setOpaque(false);
    
    JPanel rightGluePanel = new JPanel();
    rightGluePanel.setBorder(BorderFactory.createEmptyBorder());
    rightGluePanel.setPreferredSize(new Dimension(5, 5));
    rightGluePanel.setMaximumSize(rightGluePanel.getPreferredSize());
    rightGluePanel.setOpaque(false);

    JToolBar controlToolbar = new JToolBar();
    controlToolbar.setOpaque(false);
    controlToolbar.setBorder(BorderFactory.createEmptyBorder());
    controlToolbar.setFloatable(false);
    controlToolbar.setRollover(true);
    
    controlToolbar.add(leftGluePanel);
    controlToolbar.add(playButton);
    controlToolbar.add(slider);
    controlToolbar.add(timeLabel);
    controlToolbar.add(rightGluePanel);
    
    controlToolbar.setPreferredSize(new Dimension(initialWidth, TOOLBAR_HEIGHT));
    
    return controlToolbar;
  }
  
  
  private JButton createRolloverButton(Icon icon,
      Icon rolloverIcon, Icon pressedIcon, ActionListener actionListener) {

    final JButton button = new JButton(icon);
    button.setOpaque(false);
    button.setFocusPainted(false);
    button.addActionListener(actionListener);
    button.setRolloverIcon(rolloverIcon);
    button.setPressedIcon(pressedIcon);
    button.setRolloverEnabled(true);
    button.setBorderPainted(false);
    button.setContentAreaFilled(false);

    return button;
  }
  
  private void refreshTimeLabel(final long currentTimeMillis, final long totalTimeMillis) {
    final String currentTimeFormatted = millisToMinutesSeconds(currentTimeMillis);
    final String totalTimeFormatted = millisToMinutesSeconds(totalTimeMillis);
    timeLabel.setText(currentTimeFormatted + "/" + totalTimeFormatted);
  }

  private static String millisToMinutesSeconds(final long timeMillis) {
    return TimeUnit.MILLISECONDS.toMinutes(timeMillis) + ":"
        + String.format("%02d", TimeUnit.MILLISECONDS.toSeconds(timeMillis) % 60);
  }
  
  private Dimension scaleVideo(final Dimension originalVideoDimension) {
    final double scaleFactor = Math.max(originalVideoDimension.getWidth() / (double) maximumVideoDimension.getWidth(),
        originalVideoDimension.getHeight() / maximumVideoDimension.getHeight());
    final Dimension scaledVideoDimension;
    if (scaleFactor <= 1.0) {
      scaledVideoDimension = originalVideoDimension;
    } else {
      scaledVideoDimension = new Dimension((int) (originalVideoDimension.getWidth() / scaleFactor),
          (int) (originalVideoDimension.getHeight() / scaleFactor));
    }
    logger.info("Original video size: " + originalVideoDimension.width + "x" + originalVideoDimension.height
        + " scaled to " + scaledVideoDimension.width + "x" + scaledVideoDimension.height);
    return scaledVideoDimension;
  }
  
  private static boolean canConnectToVideo(final URL url){
    try {
      HttpURLConnection.setFollowRedirects(true);
      HttpURLConnection con = (HttpURLConnection) url.openConnection();
      con.setRequestMethod("HEAD");
      return (con.getResponseCode() == HttpURLConnection.HTTP_OK);
    }
    catch (Exception e) {
       e.printStackTrace();
       return false;
    }
  }
  
  public static boolean isVideoUrl(URL url) {
    if (url == null) {
      return false;
    }

    String fileName = url.getFile().toLowerCase();

    fileName = fileName.toLowerCase();
    return fileName.endsWith(".mp4");
  }

  private class VideoSurfacePanel extends JPanel {
    
    private VideoSurfacePanel() {
      setBackground(Color.black);
      setOpaque(true);
      refreshVideoSize();
    }

    public void refreshVideoSize() {
      setPreferredSize(new Dimension(videoWidth, videoHeight));
      setMinimumSize(new Dimension(videoWidth, videoHeight));
      setMaximumSize(new Dimension(videoWidth, videoHeight));
    }
    
    @Override
    protected void paintComponent(Graphics g) {
      Graphics2D g2 = (Graphics2D)g;
      g2.drawImage(image, 0, 0, videoWidth, videoHeight, null);
    }
  }
  
  private class TutorialRenderCallbackAdapter extends RenderCallbackAdapter {
    
    private TutorialRenderCallbackAdapter() {
      super(new int[maximumVideoDimension.width * maximumVideoDimension.height]);
    }
    
    @Override
    protected void onDisplay(DirectMediaPlayer mediaPlayer, int[] rgbBuffer) {
      // Simply copy buffer to the image and repaint
      image.setRGB(0, 0, maximumVideoDimension.width, maximumVideoDimension.height, rgbBuffer, 0, maximumVideoDimension.width);
      videoSurface.repaint();
    }
  }

}
